from keras.layers import Layer
import tensorflow as tf


class OnnxLSTM(Layer):
    """

    Args:
        units: int
        return_sequences: bool
        return_lstm_state: bool
        **kwargs:
    """

    def __init__(self, units: int, return_sequences: bool, return_lstm_state: bool,  **kwargs):
        super().__init__(**kwargs)
        self.lstm_layer = tf.keras.layers.LSTM(units, return_sequences=return_sequences,
                                               return_state=return_lstm_state)
        self.return_lstm_state = return_lstm_state
        self.return_sequences = return_sequences
        self.units = units

    def call(self, inputs, initial_h_state=None, initial_c_state=None, **kwargs):
        if initial_h_state is not None and initial_c_state is not None:
            initial_states = [initial_h_state, initial_c_state]
        else:
            initial_states = None
        res = self.lstm_layer(inputs, initial_state=initial_states, **kwargs)
        if self.return_lstm_state:
            lstm_tensor, h_out, c_out = res
            concat_output = tf.concat([tf.expand_dims(h_out, 1), lstm_tensor, tf.expand_dims(c_out, 1)], axis=1)
            return concat_output
        else:
            return res

    def build(self, input_shape):
        self.lstm_layer.build(input_shape)

    def get_config(self):
        config = super().get_config()
        config.update({
            "return_sequences": self.return_sequences,
            "return_lstm_state": self.return_lstm_state,
            "units": self.units
        })
        return config



import logging

import keras
import numpy as np
import tensorflow as tf
from keras import backend as K
from keras.engine.keras_tensor import KerasTensor
from keras.layers import SlicingOpLambda, Lambda
from typing import Union
from .utils import is_numpy, ensure_tf_type, unsqueeze_tensors_of_rank_one
from .tfops_funcs import tf_reshape, tf_shape, tf_cast, tf_stack, tf_image_resize, tf_strided_slice,\
    tf_squeeze, tf_transpose, tf_where, tf_gather, tf_range, tf_reduce_sum, tf_abs, tf_expand_dims, tf_concat, \
    tf_shape, tf_tile, tf_fill, tf_gather_nd, tf_reduce_sum, tf_zeros_like, tf_multiply, tf_tensor_scatter_nd_update,\
    tf_ones


def convert_transpose(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert transpose.
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    logger = logging.getLogger('onnx2keras.transpose')
    input_name = node.input[0]

    if params['perm'][0] != 0:
        logger.warning('Can\'t permute batch dimension. Result may be wrong.')
        if is_numpy(layers[input_name]):
            logger.warning('Transposing numpy array.')
            layers[node_name] = np.transpose(layers[input_name], axes=params['perm'])
        else:
            layers[node_name] = tf_transpose(layers[input_name], perm=params['perm'],
                                             tf_name=f"{params['cleaned_name']}_transpose")
    else:
        permute = keras.layers.Permute(params['perm'][1:], name=f"{params['cleaned_name']}_transpose")
        layers[node_name] = permute(layers[input_name])


def convert_shape(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert shape.
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    logger = logging.getLogger('onnx2keras.shape')
    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)

    logger.debug('Actual shape:')
    logger.debug(np.array(input_0.shape))
    is_unknown_tensor = input_0.shape == None
    if not is_unknown_tensor and (
            not K.is_keras_tensor(input_0) or not any([input_0.shape[i] == None for i in range(len(input_0.shape))])):
        shapes = []
        for i in input_0.shape:
            if i is not None:
                shapes.append(i)
            else:
                shapes.append(None)
        layers[node_name] = np.array(shapes)
    else:
        layers[node_name] = tf_shape(input_0, out_type=tf.int64, tf_name=f"{params['cleaned_name']}_shape")


def optimize_constant_array_for_serialization(input_0: tf.Tensor, params, indices: Union[np.ndarray, tf.Tensor], logger):
    """
    Some models contain repetition in the constant array tf.gather gathers from.
    In such cases, serialization takes a long time. serializing 3000 elements take ~3 sec (and scale linearly).
    To optimize for this case, we detect the repetition, take the modulu of the repetition from indices,
    and making the param array shorter.
    This is especially useful for models containing pre-top-k elements.
    Args:
        input_0: the array to gather on
        indices: the indices to gather

    Returns:
        input_0: the new array to gather on
        indices: the new indices to gather
    """
    # This Optimization is a must for models with Pre-Top-K needs
    logger.debug('onnx2keras.gather - Encountered long gather. '
                 'Trying to shorten it for easy serialization')
    max_inp = tf.reduce_max(input_0)
    min_inp = tf.reduce_min(input_0)
    if len(input_0) % (max_inp - min_inp + 1) == 0:
        range_inp = tf_range(min_inp, max_inp + 1, dtype=input_0.dtype,
                             tf_name=f"{params['cleaned_name']}_optimize_range")
        reshape_columns = tf_reshape(input_0, [-1, max_inp + 1], tf_name=f"{params['cleaned_name']}_optimize_reshape")
        if tf_reduce_sum(tf_abs((reshape_columns - range_inp[None, :]),
                                tf_name=f"{params['cleaned_name']}_optimize_abs"),
                         tf_name=f"{params['cleaned_name']}_optimize_sum") == 0:
            # This tests for a long constant array that has a repeating series like [0,1,2,3,0,1,2,3]
            logger.debug('onnx2keras.gather - Shortening sequence - columns')
            indices = indices % (max_inp + 1)
            input_0 = range_inp
        else:
            repetition_len = np.argmin(input_0 == input_0[0])
            if repetition_len > 0 and len(input_0) % repetition_len == 0:
                # This tests for a long constant array that has a repeating series like [0,0,0,1,1,1,2,2,2]
                reshaped_rows = tf_reshape(input_0, [-1, repetition_len],
                                           tf_name=f"{params['cleaned_name']}_optimize_reshape")
                first_row = reshaped_rows[:, 0]
                if tf_reduce_sum(tf_abs(reshaped_rows - first_row[:, None],
                                        tf_name=f"{params['cleaned_name']}_optimize_abs_2"),
                                 tf_name=f"{params['cleaned_name']}_optimize_sum_2") == 0:
                    logger.debug('onnx2keras.gather - Shortening sequence - rows')
                    indices = indices // repetition_len
                    input_0 = first_row
    return input_0, indices

def convert_gather(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert gather.
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    logger = logging.getLogger('onnx2keras.gather')
    OPTIMIZE_ARRAY_LENGTH = 50000
    axis = params.get('axis', 0)
    if is_numpy(layers[node.input[0]]) and is_numpy(layers[node.input[1]]) and not 'is_embedding' in params:
        logger.debug('Gather from numpy array')
        if axis == 0:
            gathered = np.array(layers[node.input[0]][layers[node.input[1]]])
        elif axis == 1:
            gathered = np.array(layers[:, node.input[0]][layers[node.input[1]]])
        elif axis == 2:
            gathered = np.array(layers[:, :, node.input[0]][layers[node.input[1]]])
        elif axis == 3:
            gathered = np.array(layers[:, :, :, node.input[0]][layers[node.input[1]]])
        else:
            raise AttributeError('Can\'t gather by axis more than 3.')

        if gathered.dtype == np.object0:
            try:
                gathered = gathered.astype(np.int32)
            except TypeError:
                pass
        layers[node_name] = gathered
    else:
        input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)
        if isinstance(layers[node.input[1]], (np.integer, int)) or \
            (not isinstance(layers[node.input[1]], np.ndarray) and \
                K.is_keras_tensor(layers[node.input[1]])):
            #indices are keras tensor or int
            indices = layers[node.input[1]]
        else: #indices are numpy/tf.eager
            indices = layers[node.input[1]]
            if not is_numpy(layers[node.input[1]]):
                indices = indices.numpy()
            indices = indices.tolist()
        if "is_embedding" in params:
            if len(input_0.shape) == 2:
                emb = tf.keras.layers.Embedding(input_0.shape[0], input_0.shape[1], weights=[layers[node.input[0]]],
                                                name=f"{params['cleaned_name']}_gather_emb")
                if isinstance(indices, list):
                    layers[node_name] = emb(np.array(indices))
                else:
                    layers[node_name] = emb(indices)
            else:
                raise AttributeError("Cannot transform gather into embedding with non 2D array")
        else:
            if tf.is_tensor(indices) and indices.dtype not in [tf.int16, tf.int32, tf.int64]:
                indices = tf_cast(indices, tf.int32, tf_name=f"{params['cleaned_name']}_gather_cast_indices")
            if isinstance(indices, list):
                indices = np.array(indices)
            if type(indices) == int:
                out_type = tf.int32
            else:
                out_type = indices.dtype
            dim_len = tf_shape(input_0, out_type=out_type,
                               tf_name=f"{params['cleaned_name']}_gather_input_shape")[axis]  # support None
            if isinstance(indices, (int, np.integer)) and indices < 0:
                indices += dim_len
            if tf.is_tensor(indices):
                indices = tf_where(indices < 0, indices + dim_len, indices,
                                   tf_name=f"{params['cleaned_name']}_gather_indices_where")
            if isinstance(input_0, np.ndarray) or not  K.is_keras_tensor(input_0):
                if len(input_0) > OPTIMIZE_ARRAY_LENGTH:
                    input_0, indices = optimize_constant_array_for_serialization(input_0, params, indices, logger)
            layers[node_name] = tf_gather(input_0, indices, axis=axis, tf_name=f"{params['cleaned_name']}_gather")


def convert_concat(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert concat.
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    logger = logging.getLogger('onnx2keras.concat')

    layer_input = [layers[node.input[i]] for i in range(len(node.input))]

    if all([is_numpy(layers[node.input[i]]) for i in range(len(node.input))]):
        logger.debug('Concat numpy arrays.')
        layers[node_name] = np.concatenate(layer_input, axis=params['axis'])
    else:
        logger.debug('Concat Keras layers.')
        if len(layer_input) > 1:
            if not np.array([tf.is_tensor(layer_input[i]) and K.is_keras_tensor(layer_input[i]) for i in
                             range(len(layer_input))]).all() or any(
                [layer_input[i].shape == None for i in range(len(layer_input))]):
                try:
                    layers[node_name] = tf_concat(layer_input, axis=params['axis'],
                                                  tf_name=f"{params['cleaned_name']}_concat")
                except Exception as ex:
                    # might be due to type mismatch between different inputs of tf.concat
                    raise

            else:
                layer_input = unsqueeze_tensors_of_rank_one(layer_input, axis=params['axis'], name=params['cleaned_name'])
                layers[node_name] = keras.layers.concatenate(inputs=layer_input,
                                                             axis=params['axis'],
                                                             name=f"{params['cleaned_name']}_concat_2")
        else:
            layers[node_name] = layer_input[0]


def convert_reshape(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert reshape.
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    logger = logging.getLogger('onnx2keras.reshape')

    input_0 = layers[node.input[0]]
    input_1 = layers[node.input[1]]

    logger.debug(f'input_0: {str(input_0)}')
    logger.debug(f'input_1: {str(input_1)}')

    if is_numpy(input_1):
        dims_to_set_as_zero = None
        dims_to_keep_unchanged = None
        allow_zero = params.get('allowzero', False)
        contains_zero_dim = np.isin(input_1, 0).any()
        contains_infer_dim = np.isin(input_1, -1).any()
        if allow_zero:
            if contains_infer_dim and contains_zero_dim:
                raise ValueError(
                    "Reshape parameter 'allowzero' is set and reshaping argument contains both '0' dim and '-1'"
                    "which is not allowed"
                    f"node name: {node_name}")
            elif contains_zero_dim:
                dims_to_set_as_zero = np.argwhere(input_1 == 0)
        elif not allow_zero and contains_zero_dim:
            dims_to_keep_unchanged = np.squeeze(np.argwhere(input_1 == 0))

        logger.debug('The second argument is numpy array.')
        if is_numpy(input_0):
            logger.debug('The first argument is numpy array. Apply np.reshape.')
            layers[node_name] = np.reshape(input_0, np.int32(input_1))
        else:
            if params['change_ordering']:
                input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)

                # Fix critical issue with NHWC
                if input_1[0] is None and input_1[1] == -1:
                    logger.warning('!!! IMPORTANT INFORMATION !!!')
                    logger.warning('The target shape if [None, -1] that means flatten.')
                    logger.warning('But the target ordering is NHWC, so we cant simply perform flatten')
                    logger.warning('The layer will be converted as lambda with tf.transpose')
                    logger.warning('---')

                    def target_layer(x):
                        import tensorflow as tf
                        x = tf.transpose(x, [0, 3, 1, 2])
                        return x

                    lambda_layer = keras.layers.Lambda(target_layer,
                                                       name="%s_CHW" % f"{params['cleaned_name']}_reshape_lambda")
                    layers[node_name] = lambda_layer(input_0)
                    lambda_func[keras_name] = target_layer
                else:
                    layers[node_name] = input_0

                reshape = keras.layers.Reshape(np.int32(input_1[1:]),
                                               name=f"{params['cleaned_name']}_reshape_input_2_1")  # keras reshape ignores batch dimension but onnx axis do not
                layers[node_name] = reshape(layers[node_name])

            else:
                input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)
                input_0_shape = input_0.shape
                if len(input_0_shape) > 0:
                    first_mismatch = np.argmin(np.array(input_0_shape[:len(input_1)]) == input_1)
                else:  # does not need to reshape dynamicalyy (skip next section)
                    first_mismatch = 0
                if (input_1 == None).any() and (np.array(input_0_shape) == None).any() and len(input_1) < len(
                        input_0_shape) \
                        and input_1[first_mismatch] == -1:  # reshape end
                    end_match_arr = np.array(input_0_shape[-len(input_1):]) == input_1
                    end_idx_match = np.argmax((np.array(input_0_shape[-len(input_1):]) == input_1))
                    end_idx_match = end_idx_match + len(input_0_shape) - len(input_1) if end_idx_match > first_mismatch \
                                                                                         and end_match_arr[
                                                                                             end_idx_match] else len(
                        input_0_shape) + 1
                    ten_shape = tf_shape(input_0, tf_name=f"{params['cleaned_name']}_shape")
                    layers[node_name] = tf_reshape(input_0,
                                                   [*ten_shape[:first_mismatch], -1,
                                                    *ten_shape[end_idx_match:]],
                                                   tf_name=f"{params['cleaned_name']}_reshape")
                else:
                    logger.debug('The first argument is Keras/tf layer. Apply keras.Reshape.')
                    logger.debug('Target shape :')
                    logger.debug(np.int32(input_1[1:]))
                    if len(np.int32(input_1[1:])) == 1 and np.int32(input_1[1:])[0] == -1:
                        if input_0.shape.rank == 1:
                            input_0 = tf_expand_dims(input_0, 0, tf_name=f"{params['cleaned_name']}_expand_dims")
                        logger.debug('The first argument is Keras/tf layer. Apply keras.Flatten.')
                        flatten = keras.layers.Reshape(target_shape=input_1[1:],
                                                       name=f"{params['cleaned_name']}_reshape_input_2_2")
                        layers[node_name] = flatten(input_0)
                    elif len(input_1) == 1 and input_1[0] == -1:
                        layers[node_name] = tf_reshape(input_0, [-1], tf_name=f"{params['cleaned_name']}_reshape_1")
                    else:
                        if len(input_0.shape) == 0 or (
                                input_0.shape[0] != input_1[0] and input_1[0] != 0):  # keras reshape don't work
                            new_shape = input_1.copy()
                            if dims_to_set_as_zero is not None:
                                new_shape[dims_to_set_as_zero] = 0
                            elif dims_to_keep_unchanged is not None:
                                new_shape[dims_to_keep_unchanged] = np.array(input_0.shape)[dims_to_keep_unchanged]
                            layers[node_name] = tf_reshape(input_0, new_shape,
                                                           tf_name=f"{params['cleaned_name']}_reshape_2")
                        else:
                            reshape = keras.layers.Reshape(np.int32(input_1[1:]),
                                                           name=f"{params['cleaned_name']}_reshape_input_2_3")
                            layers[node_name] = reshape(input_0)
    else:  # dynamic reshape
        layers[node_name] = tf_reshape(input_0, input_1, tf_name=f"{params['cleaned_name']}_reshape_3")


def convert_unsqueeze(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert unsqueeze.
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    logger = logging.getLogger('onnx2keras.unsqueeze')

    if len(node.input) != 1:
        if len(node.input) == 2:
            params['axes'] = layers[node.input[1]]
        else:
            raise AttributeError('Number of inputs is not equal 1 for unsqueeze layer')

    if len(np.unique(params['axes'])) < len(params['axes']):
        raise AttributeError(f"The specified axes contains duplicates values: {params['axes']}")

    if is_numpy(layers[node.input[0]]):
        logger.debug('Work with numpy types.')
        layers[node_name] = layers[node.input[0]]
        for axis in params['axes']:
            layers[node_name] = np.expand_dims(layers[node_name], axis)
    else:
        unsqueezed_input = layers[node.input[0]]
        for axis in params['axes']:
            unsqueezed_input = tf_expand_dims(unsqueezed_input, axis,
                                              tf_name=f"{params['cleaned_name']}_expand_dims_ax_{str(axis)}")

        layers[node_name] = unsqueezed_input


def convert_flatten(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert flatten.
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    logger = logging.getLogger('onnx2keras.flatten')

    if len(node.input) != 1:
        raise AttributeError('Number of inputs is not equal 1 for flatten layer')

    logger.debug('Convert inputs to Keras/TF layers if needed.')
    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)
    input_dims = tf_shape(input_0, tf_name=f"{params['cleaned_name']}_shape")
    flatten_axis = params.get('axis', 1)
    reshaped_input = tf_reshape(input_0, [tf.reduce_prod(input_dims[:flatten_axis]),
                                          tf.reduce_prod(input_dims[flatten_axis:])],
                                tf_name=f"{params['cleaned_name']}_flatten")
    layers[node_name] = reshaped_input


def convert_slice(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert slice.
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    logger = logging.getLogger('onnx2keras.slice')
    max_ends_val = np.iinfo(np.int32).max

    if params['change_ordering']:
        raise NotImplementedError("change_ordering for Slice is not implemented")
    if 'axes' in params:
        axes = list(params["axes"])
        ends = list(params["ends"])
        starts = list(params["starts"])
        steps = list(params.get("steps", [None] * len(axes)))
    else:
        starts = list(layers[node.input[1]])
        ends = list(layers[node.input[2]])
        try:
            axes = list(layers[node.input[3]])
        except:
            input_rank = len(layers[node.input[0]].shape)
            axes = list(range(input_rank))
        try:
            steps = list(layers[node.input[4]])
        except IndexError:
            steps = list(params.get("steps", [None] * len(axes)))

    # when the 'ends' value is the int64 maximum, probably happen because [idx:] sets large end num in conversion
    if ends[0].dtype == np.int64 and not isinstance(ends[0], KerasTensor):
        if ends[0] > max_ends_val:
            ends = [np.int32(max_ends_val)]
    try:
        max_len = len(layers[node.input[0]].shape)
        axes_positives = [axis if axis >= 0 else max_len + axis for axis in axes]
    except ValueError as e:
        if layers[node.input[0]].shape == None:  # tensor with unknown shape (not the same as dynamic)
            max_len = max(axes) + 1
            if any([axis < 0 for axis in axes]):
                raise NotImplementedError("For a tensor with unknown shape, can't use negative axis")
            else:
                axes_positives = axes
        else:
            raise NotImplementedError(f"Couldn't transform the axis in a slice layer {node_name}")
    slice_spec_param = []
    is_dynamic = False
    for i in range(len(starts)):
        for index_li in [starts, steps, ends]:
            if index_li[i] is not None and not isinstance(index_li[i], int) and not is_numpy(
                    index_li[i]) and K.is_keras_tensor(index_li[i]):
                is_dynamic = True
    if not is_dynamic:
        for axis in range(max_len):
            if axis in axes_positives:
                axis_index = axes_positives.index(axis)
                start = starts[axis_index]
                end = ends[axis_index]
                step = steps[axis_index]
                slice_spec_param.append({'start': start, 'step': step, 'stop': end})
            else:
                slice_spec_param.append({'start': None, 'step': None, 'stop': None})
        if is_numpy(layers[node.input[0]]) and np.array([_shape is None for _shape in layers[node.input[0]]]).any() \
                and len(layers[node.input[0]].shape) == 1:  # slice numpy array which is a shape
            sliced = layers[node.input[0]][start:end:step]
        else:
            input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)
            slicing_layer = SlicingOpLambda(tf.__operators__.getitem)
            sliced = slicing_layer(input_0, slice_spec=slice_spec_param)
            if is_numpy(layers[node.input[0]]) and not K.is_keras_tensor(sliced):
                sliced = sliced.numpy()
        layers[node_name] = sliced
    else:
        try:
            steps = list(layers[node.input[4]])
        except IndexError:
            steps = list(params.get("steps", [1] * len(axes)))
        input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)
        keras_shape = tf_shape(layers[node.input[0]], tf_name=f"{params['cleaned_name']}_shape")
        start_vec = [0] * max_len
        end_vec = [keras_shape[i] for i in range(max_len)]
        step_vec = [1] * max_len
        for axis in range(max_len):
            if axis in axes_positives:
                axis_index = axes_positives.index(axis)
                for res_list, input_list in zip([start_vec, step_vec, end_vec], [starts, steps, ends]):
                    slice_index = input_list[axis_index]
                    if input_list[axis_index] is not None and not isinstance(slice_index, int) and not is_numpy(
                            input_list[axis_index]) and input_list[axis_index].dtype != tf.int32:
                        slice_index = tf_cast(slice_index, tf.int32, tf_name=f"{params['cleaned_name']}_cast")
                    res_list[axis] = slice_index
        layers[node_name] = tf_strided_slice(input_0,
                                             tf.concat([start_vec], axis=0),
                                             tf.concat([end_vec], axis=0),
                                             tf.concat([step_vec], axis=0),
                                             tf_name=f"{params['cleaned_name']}_strided_slice")


def convert_squeeze(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert Squeeze layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)

    axis = None
    if 'axes' in params:
        axis = params['axes'][0]

    if len(node.input) == 2:
        axis = layers[node.input[1]].tolist()
    layers[node_name] = tf_squeeze(input_0, axis=axis, tf_name=f"{params['cleaned_name']}_squeeze")


def convert_resize(node, params, layers, lambda_func, node_name, keras_name):
    logger = logging.getLogger('onnx2keras.reshape')
    input_tensor = layers[node.input[0]]
    roi = None if len(node.input[1]) == 0 else layers[node.input[1]]
    scales = [] if len(node.input[2]) == 0 else layers[node.input[2]]
    sizes = None
    nearest_mode = params.get('nearest_mode', b'round_prefer_floor')
    if len(node.input) == 4:
        sizes = layers[node.input[3]]
    if roi:
        raise Exception("Resize with roi not supported")
    if params['mode'] == b'nearest':
        resize_method = tf.image.ResizeMethod.NEAREST_NEIGHBOR
    elif params['mode'] == b'cubic':
        resize_method = tf.image.ResizeMethod.BICUBIC
    elif params['mode'] == b'linear':
        resize_method = tf.image.ResizeMethod.BILINEAR
    else:
        raise Exception("unsupported resize method")

    rank = len(input_tensor.shape)
    axes = params.get('axes', list(range(rank)))  # Default to resizing all axes

    # Validate axes
    axes = [a if a >= 0 else a + rank for a in axes]  # Convert negative axes to positive
    if any(a < 0 or a >= rank for a in axes):  # check that all axes values are within input rank
        raise ValueError("Invalid axes value")
    to_channel_last = keras.layers.Permute((2, 3, 1),
                                           name=f"{params['cleaned_name']}_resize_channels_last")(input_tensor)  # (B, W, H, C)
    shape = tf_cast(tf_shape(input_tensor, tf_name=f"{params['cleaned_name']}_resize_shape"), tf.int32,
                    tf_name=f"{params['cleaned_name']}_resize_cast")
    if shape.shape != 4:
      raise Exception("resize layer for input tensor with rank != 4 is not supported")
    if isinstance(sizes, KerasTensor) or isinstance(scales, KerasTensor):
        tf_resize_shapes = tf_zeros_like(shape, tf_name=f"{params['cleaned_name']}_resize_zeros_like")

        if len(scales) > 0:
            for i, axis in enumerate(axes):
                indices = tf.constant([[axis]])
                update = tf_cast(tf_multiply(scales[i],
                                             tf_cast(shape[axis],
                                                     tf.float32,
                                                    tf_name=f"{params['cleaned_name']}_resize_cast_2_ax_{i}"
                                                     ),
                                             tf_name=f"{params['cleaned_name']}_resize_multiply_ax_{i}"),
                                 tf.int32,
                                 tf_name=f"{params['cleaned_name']}_resize_cast_3_ax_{i}")
                updates = tf_reshape(update, (1,), tf_name=f"{params['cleaned_name']}_resize_reshape_1_ax_{i}")
                tf_resize_shapes = tf_tensor_scatter_nd_update(tf_resize_shapes,
                                                               indices,
                                                               updates,
                                                               tf_name=f"{params['cleaned_name']}_resize_scatter_ax_{i}")

        else:
            for i, axis in enumerate(axes):
                indices = tf.constant([[axis]])
                # The value to update at the specified index
                update = tf_cast(sizes[i],
                                 tf.int32,
                                 tf_name=f"{params['cleaned_name']}_resize_cast_4_ax_{i}")
                updates = tf_reshape(update,
                                     (1,),
                                     tf_name=f"{params['cleaned_name']}_resize_reshape_1_ax_{i}")
                # Apply the update using tf.scatter_nd
                curr_name_str = f"{params['cleaned_name']}_resize_scatter_1_ax_{i}"
                tf_resize_shapes = tf_tensor_scatter_nd_update(tf_resize_shapes,
                                                               indices,
                                                               updates,
                                                               tf_name=curr_name_str)
        resize_size = tf_stack(tf_gather(tf_resize_shapes,
                                         [2, 3],
                                         tf_name=f"{params['cleaned_name']}_resize_gather"),
                               axis=0,
                               tf_name=f"{params['cleaned_name']}_resize_stack")
    else:
        tf_resize_shapes = [shape[i] for i in range(2, 4)] # (W, H) for input tensor of shape [B, C, W, H]
        if len(scales) > 0:
            for i, axis in enumerate(axes):
                if scales[i] != 1:
                    tf_resize_shapes[axis - 2] = tf_cast(scales[i] * tf_cast(tf_resize_shapes[axis - 2],
                                                                             tf.float32,
                                                                             tf_name=f"{params['cleaned_name']}_resize_cast_5_ax_{i}"),
                                                         tf.int32,
                                                         tf_name=f"{params['cleaned_name']}_resize_cast_6_ax_{i}")
        else:
            for i, axis in enumerate(axes):
                if sizes[i] != input_tensor.shape[axis]:
                    tf_resize_shapes[axis - 2] = int(sizes[i])
        resize_size = tf_stack(tf_resize_shapes,
                               axis=0,
                               tf_name=f"{params['cleaned_name']}_resize_stack_1")
    
    if (
        resize_method == tf.image.ResizeMethod.NEAREST_NEIGHBOR
        and isinstance(resize_size, keras.engine.keras_tensor.KerasTensor)
        and nearest_mode.decode() == "floor"
    ):
        logger.warning("floor nearest mode will result in faulty conversion")
    if resize_method == tf.image.ResizeMethod.NEAREST_NEIGHBOR and nearest_mode.decode() == 'floor'\
        and not isinstance(resize_size, keras.engine.keras_tensor.KerasTensor):
        if not isinstance(resize_size, np.ndarray) :
            resize_size = np.array(resize_size)

        def target_layer(x, resize_size=resize_size):
            from tensorflow.python.ops.image_ops import resize_nearest_neighbor
            
            return resize_nearest_neighbor(x, resize_size, half_pixel_centers=False)
        lambda_layer = keras.layers.Lambda(target_layer, name=f"{params['cleaned_name']}_resize_lambda")
        resized = lambda_layer(to_channel_last)
    else:
        resized = tf_image_resize(to_channel_last,
                                  size=resize_size,
                                  method=resize_method,
                                  tf_name=f"{params['cleaned_name']}_image_resize")
    to_channel_first = keras.layers.Permute((3, 1, 2),
                                            name=f"{params['cleaned_name']}_resize_channels_first")(resized)
    layers[node_name] = to_channel_first


def convert_expand(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert Expand layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    if len(node.input) != 2:
        assert AttributeError('More than 2 input for expand layer.')

    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)
    input_1 = layers[node.input[1]]
    was_converted_from_bool = False
    if input_0.dtype.is_bool:
        input_0 = tf_cast(input_0, dtype='int32', tf_name=f"{params['cleaned_name']}_bool_to_int")
        was_converted_from_bool = True
    multiply_res = input_0 * tf_ones(shape=input_1, dtype=input_0.dtype,
                                     tf_name=f"{params['cleaned_name']}_expand_use_ones")
    # input_0.dtype == np.int32 since we can't serialize constants as int64, need to cast to true type
    if layers[node.input[0]].dtype == np.int64:
        multiply_res = tf_cast(multiply_res, tf.int64, tf_name=f"{params['cleaned_name']}_to_int64")
    if was_converted_from_bool:
        multiply_res = tf_cast(multiply_res, tf.bool, tf_name=f"{params['cleaned_name']}_int_to_bool")
    layers[node_name] = multiply_res


def convert_tile(node, params, layers, lambda_func, node_name, keras_name):
    layers[node_name] = tf_tile(layers[node.input[0]], layers[node.input[1]], tf_name=f"{params['cleaned_name']}_tile")


def convert_gather_elements(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert gather.
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    logger = logging.getLogger('onnx2keras.gather_elements')
    axis = params.get('axis', 0)
    data_input = layers[node.input[0]]
    indices_input = layers[node.input[1]]

    def torch_gather(x, indices, gather_axis):
        all_indices = tf_where(tf_fill(indices.shape, True, tf_name=f"{params['cleaned_name']}_gather_fill"),
                               tf_name=f"{params['cleaned_name']}_gather_where")
        gather_locations = tf_reshape(indices, [indices.shape.num_elements()],
                                      tf_name=f"{params['cleaned_name']}_gather_reshape")

        gather_indices = []
        for axis in range(len(indices.shape)):
            if axis == gather_axis:
                gather_indices.append(tf_cast(gather_locations, dtype=tf.int64,
                                              tf_name=f"{params['cleaned_name']}_gather_cast_loc"))
            else:
                gather_indices.append(tf_cast(all_indices[:, axis], dtype=tf.int64,
                                              tf_name=f"{params['cleaned_name']}_gather_cast_all"))

        gather_indices = tf_stack(gather_indices, axis=-1, tf_name=f"{params['cleaned_name']}_gather_indices")
        gathered = tf_gather_nd(x, gather_indices, tf_name=f"{params['cleaned_name']}_gather_nd")
        reshaped = tf_reshape(gathered, indices.shape, tf_name=f"{params['cleaned_name']}_reshape")
        return reshaped

    layers[node_name] = torch_gather(data_input, indices_input, axis)



import keras
import logging
from .utils import is_numpy
from .tfops_funcs import tf_matmul, tf_shape, tf_concat, tf_reshape, tf_linalg_det, tf_linalg_matmul
import tensorflow as tf


def convert_gemm(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert Linear / GEMM layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    logger = logging.getLogger('onnx2keras.gemm')

    # Check if Bias available
    if len(node.input) == 3:
        has_bias = True
        keras_weights = [layers[node.input[1]], layers[node.input[2]]]
        logger.debug('Convert GEMM with bias.')
    elif len(node.input) == 2:
        has_bias = False
        keras_weights = [layers[node.input[1]]]
        logger.debug('Convert GEMM without bias.')
    else:
        raise AttributeError('More than 3 or less than 2 inputs')

    # Linear can have additional flag to transpose weights
    if 'transB' in params and params['transB'] == 1:
        logger.debug('Transposing W matrix.')
        keras_weights[0] = keras_weights[0].transpose()
    # Estimate input/output neurons
    input_channels, output_channels = keras_weights[0].shape[-2:]
    logger.debug('Input units %s, output units %s.', input_channels, output_channels)
    if len(layers[node.input[1]].shape) > 2: #N-dim tensor multipication Dense doesn't work
        assert len(node.input) == 2
        layers[node_name] = tf_matmul(layers[node.input[0]], layers[node.input[1]], tf_name=f"{params['cleaned_name']}_matmul")
    else:
        if is_numpy(keras_weights[0]):
            dense = keras.layers.Dense(
                output_channels,
                weights=keras_weights, name=f"{params['cleaned_name']}_gemm_dense", use_bias=has_bias
            )

            # The first input - always X
            try:
                layers[node_name] = dense(layers[node.input[0]])
            except ValueError:
                mid_shape = tf_shape(layers[node.input[0]], out_type=tf.int32, tf_name=f"{params['cleaned_name']}_shape")[:-1]
                reshape_shape = tf_concat([mid_shape, [input_channels]], axis=0, tf_name=f"{params['cleaned_name']}_concat")
                reshaped_x = tf_reshape(layers[node.input[0]], reshape_shape, tf_name=f"{params['cleaned_name']}_reshape")
                layers[node_name] = dense(reshaped_x)

        else:
            #MatMul branch should point here. If there is a bug here - split GEMM from matmul
            layers[node_name] = tf_linalg_matmul(layers[node.input[0]], layers[node.input[1]], tf_name=f"{params['cleaned_name']}_multiply")


def convert_det(node, params, layers, lambda_func, node_name, keras_name):
    layers[node_name] = tf_linalg_det(layers[node.input[0]], tf_name=f"{params['cleaned_name']}_det")



import logging

import numpy as np
import tensorflow as tf

from onnx2kerastl.customonnxlayer.onnxlstm import OnnxLSTM
from .exceptions import UnsupportedLayer
from .utils import ensure_tf_type
from .tfops_funcs import tf_cast, tf_squeeze, tf_transpose, tf_expand_dims


def convert_lstm(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert convolution layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    logger = logging.getLogger('onnx2keras.conv')

    if node.input[4] != '':
        raise UnsupportedLayer('LSTM with non default sequence_lens')
    if 'direction' in params:
        direction = params['direction']
        if isinstance(direction, bytes):
            direction = direction.decode("utf-8")
        if direction != 'forward':
            raise UnsupportedLayer(f"LSTM with {direction} direction")
    should_return_state = len(node.output) == 3
    input_tensor = tf_transpose(ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name[0]),
                                perm=[1, 0, 2],
                                tf_name=f"{params['cleaned_name']}_lstm_first_transpose")
    weights_w = layers[node.input[1]][0]
    weights_r = layers[node.input[2]][0]
    weights_b = layers[node.input[3]][0]

    initial_h_state = tf_cast(tf_squeeze(ensure_tf_type(layers[node.input[5]]),
                                         axis=0,
                                         tf_name=f"{params['cleaned_name']}_lstm_squeeze_h"
                                         ),
                              input_tensor.dtype,
                              tf_name=f"{params['cleaned_name']}_lstm_cast_h")
    initial_c_state = tf_cast(
        tf_squeeze(
        ensure_tf_type(layers[node.input[6]]),
        axis=0,
        tf_name=f"{params['cleaned_name']}_lstm_squeeze_c"), input_tensor.dtype,
        tf_name=f"{params['cleaned_name']}_lstm_cast_c")

    tf.keras.backend.set_image_data_format("channels_last")
    hidden_size = params['hidden_size']
    lstm_layer = OnnxLSTM(hidden_size, return_sequences=True, return_lstm_state=should_return_state)
    res = lstm_layer(input_tensor, initial_h_state, initial_c_state)
    # prepare the keras lstm weights from the onnx inputs:
    w1 = np.concatenate([weights_w[0:hidden_size, :], weights_w[2 * hidden_size:3 * hidden_size, :],
                         weights_w[3 * hidden_size:4 * hidden_size, :],
                         weights_w[hidden_size:2 * hidden_size, :]]).transpose()
    w2 = np.concatenate([weights_r[0:hidden_size, :], weights_r[2 * hidden_size:3 * hidden_size, :],
                         weights_r[3 * hidden_size:4 * hidden_size, :],
                         weights_r[hidden_size:2 * hidden_size, :]]).transpose()
    weights_b_part1 = weights_b[:w2.shape[1]]
    weights_b_part2 = weights_b[w2.shape[1]:]
    bias1 = np.concatenate([weights_b_part1[0:hidden_size], weights_b_part1[2 * hidden_size:3 * hidden_size],
                            weights_b_part1[3 * hidden_size:4 * hidden_size],
                            weights_b_part1[hidden_size:2 * hidden_size]]).transpose()
    bias2 = np.concatenate([weights_b_part2[0:hidden_size], weights_b_part2[2 * hidden_size:3 * hidden_size],
                            weights_b_part2[3 * hidden_size:4 * hidden_size],
                            weights_b_part2[hidden_size:2 * hidden_size]]).transpose()
    bias = bias1 + bias2
    res.node.layer.set_weights([w1, w2, bias])
    tf.keras.backend.set_image_data_format("channels_first")
    if should_return_state:
        c_out = res[:, -1, :]
        h_out = res[:, 0, :]

        # the shapes of the hidden and cell should be [num_directions, batch_size, hidden_size]
        # for now we support only direction=forward so num_direction = 1 and we add directions dimension,
        # if we support direction=bidirectional we should handle it well in the lstm layer and probably remove the
        # expand dims here
        c_out = tf.expand_dims(c_out, 0)
        h_out = tf.expand_dims(h_out, 0)

        lstm_tensor = res[:, 1:-1, :]
        layers[node.output[1]] = h_out
        layers[node.output[2]] = c_out
    else:
        lstm_tensor = res
    lstm_tensor_in_onnx_order = tf_transpose(lstm_tensor, perm=[1, 0, 2], tf_name=f"{params['cleaned_name']}_lstm_transpose")
    lstm_tensor_in_onnx_order = tf_expand_dims(lstm_tensor_in_onnx_order, axis=1,
                                               tf_name=f"{params['cleaned_name']}_lstm_expand_dims")
    layers[node_name] = lstm_tensor_in_onnx_order

def convert_gru(node, params, layers, lambda_func, node_name, keras_name):
    logger = logging.getLogger('onnx2keras.convert_gru')
    if len(params["_outputs"]) > 1:
        logger.warning("The GRU return hidden state is currently not supported. Accessing in deeper layers will raise Exception")
    if params.get('activation_alpha') or params.get('activation_beta') or params.get('activations'):
        raise NotImplementedError('Custom Activations in GRU not implemented')
    if params.get('clip'):
        raise NotImplementedError('Clip in GRU not implemented')
    if params.get('direction'):  #After implementation - verify weights reshaping, and h default_size for all directions
        raise NotImplementedError('direction in  GRU not implemented')
    else:
        num_directions = 1
    if params.get('layout'):
        raise NotImplementedError('GRU layout not supported (currently supporting opset 7)')
    else:
        layout = 0
    if node.input[4] != "":
        raise NotImplementedError('GRU sequence_lens is not yet implemented')
    hidden_size = params.get('hidden_size')
    linear_before_reset = bool(params.get('linear_before_reset', 0))
    x = layers[node.input[0]] # [seq_length, batch_size, input_size] iff layout = 0
    w = layers[node.input[1]]
    r = layers[node.input[2]]
    b = layers.get(node.input[3], np.zeros((num_directions, 6*hidden_size), dtype=np.float32))
    h = layers.get(node.input[5], np.zeros((1, x.shape[1], hidden_size), dtype=np.float32))
    if isinstance(h, np.ndarray):
        tensor_h = tf.convert_to_tensor(h)
    else:
        tensor_h = h
    tf.keras.backend.set_image_data_format("channels_last")
    gru_layer = tf.keras.layers.GRU(units=hidden_size,
                        reset_after=linear_before_reset,
                        return_sequences=True,
                        name=f"{params['cleaned_name']}_gru")
    if layout == 0:
        batch_first_x = tf_transpose(x, [1, 0, 2], tf_name=f"{params['cleaned_name']}_gru_transpose")
    res = gru_layer(batch_first_x, initial_state=tf.convert_to_tensor(tensor_h[0]))
    # gru_layer.build(tf.shape(batch_first_x))
    gru_layer.set_weights([w[0].swapaxes(0, 1), r[0].swapaxes(0, 1), b[0].reshape(-1, 3*hidden_size)])
    # res = gru_layer(batch_first_x, initial_state=tf.convert_to_tensor(tensor_h[0]))
    if num_directions == 1:
        reshaped_res = tf_expand_dims(tf_transpose(res,
                                                   [1, 0, 2],
                                                   tf_name=f"{params['cleaned_name']}_gru_transpose"),
                                      axis=1,
                                      tf_name=f"{params['cleaned_name']}")
    else:
        raise NotImplementedError("GRU bidirectional output reshaping is not implemented")
    layers[node_name] = reshaped_res
    tf.keras.backend.set_image_data_format("channels_first")



import logging
from functools import partial
from typing import List

import keras
import numpy as np
import tensorflow as tf
from tensorflow.python.framework.ops import EagerTensor

from .utils import ensure_tf_type, is_numpy
from .tfops_funcs import tf_transpose, tf_pad, tf_shape, tf_reshape

def calculate_permute_values(n_dims: int, to_channel_first: bool) -> List[int]:
    if to_channel_first:
        return [n_dims - 1] + list(range(1, n_dims - 1))
    else:
        return list(range(2, n_dims)) + [1]


def permute_wrap_conv_if_constant(partial_func, conv_input, is_constant, conv_channels, params):
    if is_constant:
        input_shape = tf_shape(conv_input, tf_name=f"{params['cleaned_name']}_conv_wrap_shape")
        permuted = keras.layers.Permute(calculate_permute_values(len(input_shape), to_channel_first=False),
                                        name=f"{params['cleaned_name']}_conv_wrap_permute_1")(conv_input)
        conv_res = partial_func(data_format="channels_last")(permuted)
        result = keras.layers.Permute(calculate_permute_values(len(input_shape), to_channel_first=True),
                                      name=f"{params['cleaned_name']}_conv_wrap_permute_2")(conv_res)
    else:
        data_fmt = keras.backend.image_data_format()
        conv = partial_func(data_format=data_fmt)
        if data_fmt == 'channels_first':
            channels_idx = 1
        else:
            channels_idx = -1
        if conv_input.shape[channels_idx] is None:  # This will not serialize well unless we reshape input
            conv_input_shape = tf_shape(conv_input, tf_name=f"{params['cleaned_name']}_conv_wrap_shape_1")
            conv_input = tf_reshape(conv_input, [*conv_input_shape[:channels_idx], conv_channels,
                                                 *conv_input_shape[channels_idx + 1:]],
                                    tf_name=f"{params['cleaned_name']}_conv_wrap_reshape_2")
        if conv_input.shape[-1] is None:
            conv.build((None, conv_channels, *conv_input.shape[2:]))
        result = conv(conv_input)
    return result


def convert_conv(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert convolution layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    logger = logging.getLogger('onnx2keras.conv')

    if len(node.input) == 3:
        logger.debug('Conv with bias')
        # Has bias
        has_bias = True
        W = layers[node.input[1]]
        bias = layers[node.input[2]]

    elif len(node.input) == 2:
        logger.debug('Conv without bias')
        has_bias = False
        W = layers[node.input[1]]
        bias = None

    else:
        raise NotImplementedError('Not implemented')

    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)
    is_constant = is_numpy(input_0) or isinstance(input_0, EagerTensor)
    is_W_constant = is_numpy(W) or isinstance(W, EagerTensor)
    n_groups = params['group'] if 'group' in params else 1
    dilation = params['dilations'][0] if 'dilations' in params else 1
    pads = params['pads'] if 'pads' in params else [0, 0, 0]
    strides = params['strides'] if 'strides' in params else [1, 1, 1]

    if len(W.shape) == 5:  # 3D conv
        logger.debug('3D convolution')
        if pads[0] > 0 or pads[1] > 0 or pads[2] > 0:
            logger.debug('Paddings exist, add ZeroPadding layer')
            padding_name = f"{params['cleaned_name']}_" + 'conv_pad'
            padding_layer = keras.layers.ZeroPadding3D(
                padding=(pads[0], pads[1], pads[2]),
                name=padding_name
            )
            layers[padding_name] = input_0 = padding_layer(input_0)
        out_channels, channels_per_group, dimension, height, width = W.shape
        W = W.transpose(2, 3, 4, 1, 0)

        if has_bias:
            weights = [W, bias]
        else:
            weights = [W]
        conv_args = {"filters": out_channels,
                     "kernel_size": (dimension, height, width),
                     "strides": (strides[0], strides[1], strides[2]),
                     "padding": 'valid',
                     "weights": weights,
                     "use_bias": has_bias,
                     "activation": None,
                     "dilation_rate": dilation,
                     "name": f"{params['cleaned_name']}_" + 'conv',
                     "groups": n_groups}
        partial_conv = partial(keras.layers.Conv3D, **conv_args)
        layers[node_name] = permute_wrap_conv_if_constant(partial_conv, input_0, is_constant, weights[0].shape[-2]*n_groups, params)

    elif len(W.shape) == 4:  # 2D conv
        logger.debug('2D convolution')

        padding = None
        if len(pads) == 2 and (pads[0] > 0 or pads[1] > 0):
            padding = (pads[0], pads[1])
        elif len(pads) == 4 and (pads[0] > 0 or pads[1] > 0 or pads[2] > 0 or pads[3] > 0):
            padding = ((pads[0], pads[2]), (pads[1], pads[3]))

        if padding:
            logger.debug('Paddings exist, add ZeroPadding layer')
            padding_name = f"{params['cleaned_name']}_" + 'conv_pad_1'
            padding_layer = keras.layers.ZeroPadding2D(
                padding=padding,
                name=padding_name,
                data_format='channels_first'
            )
            layers[padding_name] = input_0 = padding_layer(input_0)

        W = W.transpose(2, 3, 1, 0) if is_W_constant else tf.transpose(W, [2, 3, 1, 0])
        height, width, channels_per_group, out_channels = W.shape

        if has_bias:
            weights = [W, bias]
        else:
            weights = [W]
        if is_W_constant:
            conv_args = {"filters": out_channels,
                         "kernel_size": (height, width),
                         "strides": (strides[0], strides[1]),
                         "padding": 'valid',
                         "weights": weights,
                         "use_bias": has_bias,
                         "activation": None,
                         "dilation_rate": dilation,
                         "name": f"{params['cleaned_name']}_" + 'conv',
                         "groups": n_groups}

            partial_conv = partial(keras.layers.Conv2D, **conv_args)
            layers[node_name] = permute_wrap_conv_if_constant(partial_conv, input_0, is_constant, weights[0].shape[-2]*n_groups, params)
        else:
            input_0_nhwc = tf_transpose(input_0, [0, 2, 3, 1],
                                        tf_name=f"{params['cleaned_name']}_" + 'conv_transpose_nhwc')

            # Perform the convolution in NHWC format
            conv_nhwc = tf.nn.conv2d(input_0_nhwc, weights[0], strides=(strides[0], strides[1]),
                                     dilations=dilation,
                                     padding='VALID', data_format='NHWC',
                                     name=f"{params['cleaned_name']}_" + 'conv')

            # Permute the result back to NCHW format
            layers[node_name] = tf_transpose(conv_nhwc, [0, 3, 1, 2],
                                             tf_name=f"{params['cleaned_name']}_" + 'conv_transpose_2_nchw')
    else:
        # 1D conv
        W = W.transpose(2, 1, 0)
        width, channels, n_filters = W.shape
        print(width, channels, n_filters, has_bias)

        weights = [W]
        conv_args = {"filters": n_filters,
                     "kernel_size": (width),
                     "strides": (strides[0]),
                     "weights": weights,
                     "use_bias": False,
                     "activation": None,
                     "dilation_rate": dilation,
                     "name": f"{params['cleaned_name']}_" + 'conv',
                     "groups": n_groups}

        padding = None
        if len(pads) == 2 and (pads[0] > 0 or pads[1] > 0):
            padding = (pads[0], pads[1])

        if padding:
            # find the dimension to pad and use the exact padding values
            input_shape = np.asarray(keras.backend.int_shape(input_0))
            partitioned_dim = np.argwhere(input_shape == channels * n_groups)[0][0]
            padding_dim = 2 if partitioned_dim == 1 else 1
            tf_padding = np.zeros((2, len(input_shape))).astype(int)
            tf_padding[:, padding_dim] = [padding[0], padding[1]]
            input_0 = tf_pad(input_0, tf.constant(list(tf_padding.transpose())),
                             tf_name=f"{params['cleaned_name']}_conv_pad_0")
        else:
            conv_args['padding'] = 'valid'
        partial_conv = partial(keras.layers.Conv1D, **conv_args)
        res = permute_wrap_conv_if_constant(partial_conv, input_0, is_constant, weights[0].shape[-2]*n_groups, params)
        if has_bias:
            res_shape = np.asarray(keras.backend.int_shape(res))
            bias_dim = np.argwhere(res_shape == bias.shape)[0][0]
            expanded_dims = [dim for dim in range(len(res_shape)) if dim != bias_dim]
            res = res + np.expand_dims(bias, expanded_dims)

        layers[node_name] = res


def convert_convtranspose(node, params, layers,
                          lambda_func, node_name, keras_name):
    """
    Convert transposed convolution layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    logger = logging.getLogger('onnx2keras.convtranpose')

    if len(node.input) == 3:
        logger.debug('ConvTranspose with bias')
        # Has bias
        has_bias = True
        W = layers[node.input[1]]
        bias = layers[node.input[2]]

    elif len(node.input) == 2:
        logger.debug('ConvTranspose without bias')
        has_bias = False
        W = layers[node.input[1]]
        bias = None

    else:
        raise NotImplementedError('Not implemented')

    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)
    is_W_constant = is_numpy(W) or isinstance(W, EagerTensor)
    n_groups = params['group'] if 'group' in params else 1
    dilation = params['dilations'][0] if 'dilations' in params else 1
    pads = params['pads'] if 'pads' in params else [0, 0]
    strides = params['strides'] if 'strides' in params else [1, 1]

    if len(W.shape) == 5:  # 3D conv
        W = W.transpose(2, 3, 4, 1, 0)
        height, width, depth, n_filters, channels = W.shape

        if has_bias:
            weights = [W, bias]
        else:
            weights = [W]

        if n_groups > 1:
            raise AttributeError('Cannot convert ConvTranspose2d with groups != 1')

        if dilation > 1:
            raise AttributeError('Cannot convert ConvTranspose2d with dilation_rate != 1')

        conv = keras.layers.Conv3DTranspose(
            filters=n_filters,
            kernel_size=(height, width, depth),
            strides=strides,
            padding='valid',
            output_padding=0,
            weights=weights,
            use_bias=has_bias,
            activation=None,
            dilation_rate=dilation,
            name=f"{params['cleaned_name']}_convtranspose"
        )

        if 'output_shape' in params and 'pads' not in params:
            logger.debug('!!!!! Paddings will be calculated automatically !!!!!')
            pads = [strides[0] * (int(input_0.shape[2]) - 1) + 0 + (height - 1) * dilation - params['output_shape'][0],
                    strides[1] * (int(input_0.shape[3]) - 1) + 0 + (height - 1) * dilation - params['output_shape'][1]]

        layers[node_name] = input_0 = conv(input_0)

        # Magic ad-hoc.
        # See the Keras issue: https://github.com/keras-team/keras/issues/6777
        # input_0.set_shape(input_0.shape)

        if 'output_padding' in params and (params['output_padding'][0] > 0 or params['output_padding'][1] > 0):
            raise AttributeError('Cannot convert ConvTranspose2d with output_padding != 0')

        if pads[0] > 0:
            logger.debug('Add cropping layer for output padding')
            assert (len(pads) == 2 or (pads[2] == pads[0] and pads[3] == pads[1]))

            crop = keras.layers.Cropping2D(
                pads[:2],
                name=f"{params['cleaned_name']}_convtranspose" + '_crop'
            )
            layers[node_name] = crop(input_0)

    elif len(W.shape) == 4:  # 2D conv
        W = W.transpose(2, 3, 1, 0) if is_W_constant else tf.transpose(W, [2, 3, 1, 0])
        height, width, n_filters, channels = W.shape

        if has_bias:
            weights = [W, bias]
        else:
            weights = [W]

        if n_groups > 1:
            raise AttributeError('Cannot convert ConvTranspose2d with groups != 1')

        if dilation > 1:
            raise AttributeError('Cannot convert ConvTranspose2d with dilation_rate != 1')
        if is_W_constant:
            conv = keras.layers.Conv2DTranspose(
                filters=n_filters,
                kernel_size=(height, width),
                strides=strides,
                padding='valid',
                output_padding=0,
                weights=weights,
                use_bias=has_bias,
                activation=None,
                dilation_rate=dilation,
                name=f"{params['cleaned_name']}_convtranspose"
            )

            if 'output_shape' in params and 'pads' not in params:
                logger.debug('!!!!! Paddings will be calculated automatically !!!!!')
                pads = [
                    strides[0] * (int(input_0.shape[2]) - 1) + 0 + (height - 1) * dilation - params['output_shape'][0],
                    strides[1] * (int(input_0.shape[3]) - 1) + 0 + (height - 1) * dilation - params['output_shape'][1]]

            layers[node_name] = input_0 = conv(input_0)
        else:
            input_0_nhwc = tf.transpose(input_0, [0, 2, 3, 1])

            output_shape = infer_output_shape(input_shape=tf.shape(input_0_nhwc), filter_shape=tf.shape(W),
                                              strides=strides,
                                              padding='VALID')
            conv_transpose_nhwc = tf.nn.conv2d_transpose(input_0_nhwc, weights[0], output_shape=output_shape,
                                                         strides=(strides[0], strides[1]), dilations=dilation, padding='VALID',
                                                         data_format='NHWC',
                                                         name=f"{params['cleaned_name']}_convtranspose_nhwc")

            # Permute the result back to NCHW format
            layers[node_name] = tf_transpose(conv_transpose_nhwc, [0, 3, 1, 2],
                                             tf_name=f"{params['cleaned_name']}_convtranspose")

        # Magic ad-hoc.
        # See the Keras issue: https://github.com/keras-team/keras/issues/6777
        # input_0.set_shape(input_0.shape)

        if 'output_padding' in params and (params['output_padding'][0] > 0 or params['output_padding'][1] > 0):
            raise AttributeError('Cannot convert ConvTranspose2d with output_padding != 0')

        if pads[0] > 0:
            logger.debug('Add cropping layer for output padding')
            assert (len(pads) == 2 or (pads[2] == pads[0] and pads[3] == pads[1]))

            crop = keras.layers.Cropping2D(
                pads[:2],
                name=f"{params['cleaned_name']}_convtranspose" + '_crop_1'
            )
            layers[node_name] = crop(input_0)
    else:
        raise AttributeError('Layer is not supported for now')


def infer_output_shape(input_shape, filter_shape, strides, padding):
    input_size_h, input_size_w = input_shape[1:3]
    filter_size_h, filter_size_w = filter_shape[0:2]

    if padding == 'SAME':
        pad_h = max((input_size_h - 1) * strides[0] + filter_size_h - input_size_h, 0) // 2
        pad_w = max((input_size_w - 1) * strides[1] + filter_size_w - input_size_w, 0) // 2
    elif padding == 'VALID':
        pad_h = 0
        pad_w = 0
    else:
        raise ValueError("Padding must be 'SAME' or 'VALID'")

    output_size_h = (input_size_h - 1) * strides[0] + filter_size_h - 2 * pad_h
    output_size_w = (input_size_w - 1) * strides[1] + filter_size_w - 2 * pad_w

    return [input_shape[0], output_size_h,
            output_size_w, filter_shape[2]]  # [batch_size, output_channels, output_height, output_width]



from typing import List, Union, Callable

import numpy as np
import keras
from keras.engine.keras_tensor import KerasTensor
from keras_data_format_converter import convert_channels_first_to_last
import tensorflow as tf
from.tfops_funcs import tf_reshape

ONNX_ELEM_TO_TF_TYPE = {
    1: tf.float32,
    2: tf.uint8,
    3: tf.int8,
    4: tf.uint16,
    5: tf.int16,
    6: tf.int32,
    7: tf.int64,
    8: tf.string,
    9: tf.bool,
    10: tf.float16,
    11: tf.double,
    12: tf.uint32,
    13: tf.uint64,
    14: tf.complex64,
    15: tf.complex128,
    16: tf.bfloat16
}


def is_numpy(obj):
    """
    Check of the type is instance of numpy array
    :param obj: object to check
    :return: True if the object is numpy-type array.
    """
    return isinstance(obj, (np.ndarray, np.generic))


def ensure_tf_type(obj, name="Const"):
    import numpy as np
    import tensorflow as tf
    """
    Convert to Keras Constant if needed
    :param obj: numpy / tf type
    :param fake_input_layer: fake input layer to add constant
    :return: tf type
    """
    if is_numpy(obj): # TF < v1.16 assumes all ints are int32 and all floats are float32
        if obj.dtype == np.int64:
            obj = np.int32(obj)
        return tf.constant(obj, name=name)
    else:
        return obj


def check_torch_keras_error(model, k_model, input_np, epsilon=1e-5, change_ordering=False,
                            should_transform_inputs=False):
    """
    Check difference between Torch and Keras models
    :param model: torch model
    :param k_model: keras model
    :param input_np: input data as numpy array or list of numpy array
    :param epsilon: allowed difference
    :param change_ordering: change ordering for keras input
    :param should_transform_inputs: default False, set to True for converting channel first inputs to  channel last format
    :return: actual difference

    """
    from torch.autograd import Variable
    import torch

    if isinstance(input_np, np.ndarray):
        input_np = [input_np.astype(np.float32)]

    input_var = [Variable(torch.FloatTensor(i)) for i in input_np]
    pytorch_output = model(*input_var)
    if isinstance(pytorch_output, dict):
        pytorch_output = [p.data.numpy() for p in list(pytorch_output.values())]
    elif isinstance(pytorch_output, (tuple, list)):
        pytorch_output = [p.data.numpy() for p in pytorch_output]
    else:
        pytorch_output = [pytorch_output.data.numpy()]

    if change_ordering:
        # change image data format

        # to proper work with Lambda layers that transpose weights based on image_data_format
        keras.backend.set_image_data_format("channels_last")

        _input_np = []
        for i in input_np:
            axes = list(range(len(i.shape)))
            axes = axes[0:1] + axes[2:] + axes[1:2]
            _input_np.append(np.transpose(i, axes))
        input_np = _input_np

        # run keras model
        keras_output = k_model.predict(input_np)
        if not isinstance(keras_output, list):
            keras_output = [keras_output]

        # change image data format if output shapes are different (e.g. the same for global_avgpool2d)
        _koutput = []
        for i, k in enumerate(keras_output):
            if k.shape != pytorch_output[i].shape:
                axes = list(range(len(k.shape)))
                axes = axes[0:1] + axes[-1:] + axes[1:-1]
                k = np.transpose(k, axes)
            _koutput.append(k)
        keras_output = _koutput
    else:
        inputs_to_transpose = []
        if should_transform_inputs:
            inputs_to_transpose = [k_input.name for k_input in k_model.inputs]

            _input_np = []
            for i in input_np:
                axes = list(range(len(i.shape)))
                axes = axes[0:1] + axes[2:] + axes[1:2]
                _input_np.append(np.transpose(i, axes))
            input_np = _input_np

        k_model = convert_channels_first_to_last(k_model, inputs_to_transpose)
        if len(input_np) == 1:
            input_np = input_np[0]
        keras_output = k_model(input_np)
        if not isinstance(keras_output, list):
            keras_output = [keras_output]

        _koutput = []
        for i, k in enumerate(keras_output):
            if k.shape != pytorch_output[i].shape:
                axes = list(range(len(k.shape)))
                axes = axes[0:1] + axes[-1:] + axes[1:-1]
                k = np.transpose(k, axes)
            _koutput.append(k)
        keras_output = _koutput

    max_error = 0
    for p, k in zip(pytorch_output, keras_output):
        error = np.max(np.abs(p - k))
        np.testing.assert_allclose(p, k, atol=epsilon, rtol=0.0)
        if error > max_error:
            max_error = error

    return max_error


def unsqueeze_tensors_of_rank_one(tensor_list, axis: int, name: str):
    """
    Adjusts the ranks of tensors of rank 1 in a given list to match the maximum rank by adding dummy dimensions
    e.g., for input tensors shapes [(2,), (1, 4)] the unsqueezed tensors are [(1, 2), (1, 4)]

    Args:
        tensor_list (list): A list of tensors.

    Returns:
        list: A new list of tensors with adjusted ranks to match the maximum rank.
              If all tensors in the input list already have the same rank, the original list is returned.
    """
    ranks = [tensor.shape.rank for tensor in tensor_list]
    max_rank = max(ranks)

    if len(set(ranks)) == 1:
        return tensor_list
    elif len(set(ranks)) > 2:
        raise ValueError(f"More than 2 different ranks detected, broadcasting is ambiguous.\n"
                         f"Check the outputs of layers: \n" + '\n'.join([tensor.name for tensor in tensor_list]))

    unsqueezed_tensors = []
    for tensor in tensor_list:
        tensor_rank = tensor.shape.rank
        if tensor_rank == 1:
            rank_diff = max_rank - 1
            new_shape = [1] * axis + list(tensor.shape) + [1] * (rank_diff - axis)
            unsqueezed_tensor = tf_reshape(tensor, new_shape, tf_name=f"{name}_rank_one_reshape")
            unsqueezed_tensors.append(unsqueezed_tensor)
        else:
            unsqueezed_tensors.append(tensor)

    return unsqueezed_tensors


def ensure_float(value):
    if isinstance(value, (list, np.ndarray)):
        return float(value[0])
    elif isinstance(value, tf.Tensor):
        return float(value.numpy().item())
    else:
        return float(value)



import keras
import logging
from .utils import ensure_tf_type
from .utils import is_numpy
from .tfops_funcs import tf_pad
import tensorflow as tf


def convert_padding(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert Constant layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    # It's binary by-default
    logger = logging.getLogger("onnx2keras.padding")
    if 'mode' in params:
        params['mode'] = params['mode'].decode('ascii')
    else:
        params['mode'] = 'constant'
    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)

    if 'pads' in params:
        pads = params['pads']
    else:
        pads = layers[node.input[1]]

    if (is_numpy(pads) or not keras.backend.is_keras_tensor(pads)) and not any(pads):
        layers[node_name] = input_0
        return

    if params['mode'] == 'constant':

        if 'value' in params and params['value'] != 0.0:
            raise AssertionError('Cannot convert non-zero padding')
        if pads.shape[0] == 6 and len(layers[node.input[0]].shape) == 3:
            layers[node_name] = tf_pad(input_0, [[pads[0], pads[3]], [pads[1], pads[4]], [pads[2], pads[5]]],
                                       tf_name=f"{params['cleaned_name']}_pad")
        # Magic ordering
        else:
            if isinstance(pads, keras.engine.keras_tensor.KerasTensor) and pads.shape[0] == 8:
                padding_layer = lambda x: tf_pad(x,
                                                 [[pads[0], pads[4]],
                                                     [pads[1], pads[5]],
                                                     [pads[2], pads[6]],
                                                     [pads[3], pads[7]]],
                                                 tf_name=f"{params['cleaned_name']}_pad_3")
            elif pads.shape[0] == 8:
                padding_layer = keras.layers.ZeroPadding2D(
                    padding=((pads[2], pads[6]), (pads[3], pads[7])),
                    name=f"{params['cleaned_name']}_pad_0"
                )
            elif pads.shape[0] == 12:  # Check for rank 6 input
                padding_layer = keras.layers.Lambda(
                    lambda x: tf.pad(
                        x,
                        [
                            [0, 0],  # Batch dimension
                            [0, 0],  # Channels dimension
                            [pads[2], pads[8]],  # d1 dimension
                            [pads[3], pads[9]],  # d2 dimension
                            [pads[4], pads[10]],  # d3 dimension
                            [pads[5], pads[11]],  # d4 dimension
                        ],
                        mode='CONSTANT'
                    ),
                    name=f"{params['cleaned_name']}_pad_1"
                )
                layers[node_name] = padding_layer(input_0)
            else:
                logger.warning("Caution - no test yet")
                padding_layer = keras.layers.ZeroPadding3D(
                    padding=((pads[2], pads[7]), (pads[3], pads[8]), (pads[4], pads[9])),
                    name=f"{params['cleaned_name']}_pad_2"
                )
            layers[node_name] = padding_layer(input_0)
    elif params['mode'] == 'reflect':
        if pads.shape[0] == 6:
            result = tf_pad(input_0, [[pads[0], pads[3]], [pads[1], pads[4]], [pads[2], pads[5]]], mode='REFLECT',
                                       tf_name=f"{params['cleaned_name']}_reflect_pad")
            layers[node_name] = result
        else:
            def target_layer(x, pads=pads):
                if pads.shape[0] == 8:
                    layer = tf.pad(x, [[0, 0], [0, 0], [pads[2], pads[6]], [pads[3], pads[7]]], 'REFLECT')
                else:
                    logger.warning("Caution - no test yet")
                    layer = tf.pad(x, [[0, 0], [0, 0], [pads[2], pads[7]], [pads[3], pads[8]], [pads[4], pads[9]]], 'REFLECT')
                return layer

            lambda_layer = keras.layers.Lambda(target_layer, name=f"{params['cleaned_name']}_pad_reflect")
            layers[node_name] = lambda_layer(input_0)
            lambda_func[keras_name] = target_layer
    elif params['mode'] == 'edge':

        def target_layer(x, pads=pads):
            import tensorflow as tf
            if pads.shape[0] == 8:  # TODO not tested yet
                layer = tf.pad(x, [[0, 0], [0, 0], [pads[2], pads[6]], [pads[3], pads[7]]], 'SYMMETRIC')
            else:
                logger.warning("Caution - no test yet")
                layer = tf.pad(x, [[0, 0], [0, 0], [pads[2], pads[7]], [pads[3], pads[8]], [pads[4], pads[9]]], 'SYMMETRIC')
            return layer

        lambda_layer = keras.layers.Lambda(target_layer, name=f"{params['cleaned_name']}_pad_edge")
        layers[node_name] = lambda_layer(input_0)
        lambda_func[keras_name] = target_layer

    else:
        raise AttributeError('Unknown padding')



"""
The ONNX to keras converter module
"""
import importlib.util
import inspect
import logging
import uuid
from dataclasses import dataclass
from typing import Optional

import keras
import keras.backend
import tensorflow as tf
from keras.models import Model

from .customonnxlayer import onnx_custom_objects_map
from .exceptions import UnsupportedLayer, OnnxUnsupported
from .layers import AVAILABLE_CONVERTERS
import re

onnx_imported = False
package_name = 'onnx'
spec = importlib.util.find_spec(package_name)
if spec is not None:
    from onnx import numpy_helper

    onnx_imported = True


@dataclass
class ConvertedResponse:
    converted_model: Model
    error_info: Optional[str] = None


def onnx_node_attributes_to_dict(args):
    """
    Parse ONNX attributes to Python dictionary
    :param args: ONNX attributes object
    :return: Python dictionary
    """

    def onnx_attribute_to_dict(onnx_attr):
        """
        Parse ONNX attribute
        :param onnx_attr: ONNX attribute
        :return: Python data type
        """
        if onnx_attr.HasField('t'):
            return numpy_helper.to_array(getattr(onnx_attr, 't'))

        for attr_type in ['f', 'i', 's']:
            if onnx_attr.HasField(attr_type):
                return getattr(onnx_attr, attr_type)

        for attr_type in ['floats', 'ints', 'strings']:
            if getattr(onnx_attr, attr_type):
                return list(getattr(onnx_attr, attr_type))

    return {arg.name: onnx_attribute_to_dict(arg) for arg in args}


def flatten_onnx_nodes(onnx_nodes):
    onnx_list = []
    for node in onnx_nodes:
        if node.op_type == "If":
            onnx_list += flatten_onnx_nodes(node.attribute[0].g.node)
            onnx_list += flatten_onnx_nodes(node.attribute[1].g.node)
        onnx_list.append(node)
    return onnx_list


def onnx_to_keras(onnx_model, input_names, name_policy=None, verbose=True, change_ordering=False, input_types=None,
                  allow_partial_compilation=True) \
        -> ConvertedResponse:
    """
    Convert ONNX graph to Keras model format
    :param onnx_model: loaded ONNX model
    :param input_names: list with input names
    :param name_policy: override layer names. None, "short", "renumerate" or "attach_weights_name" (last 2 are experimental)
    :param verbose: verbose output
    :param change_ordering: change ordering to HWC (experimental)
    :return: Keras model
    """
    if not onnx_imported:
        raise OnnxUnsupported()
    # Use channels first format by default.
    keras_fmt = keras.backend.image_data_format()
    keras.backend.set_image_data_format('channels_first')

    if verbose:
        logging.basicConfig(level=logging.DEBUG)

    logger = logging.getLogger('onnx2keras')

    logger.info('Converter is called.')

    onnx_weights = onnx_model.graph.initializer
    onnx_inputs = onnx_model.graph.input
    onnx_outputs = [i.name for i in onnx_model.graph.output]
    onnx_nodes = flatten_onnx_nodes(onnx_model.graph.node) # Pulls graph out of If nodes
    logger.debug('List inputs:')
    for i, input in enumerate(onnx_inputs):
        logger.debug('Input {0} -> {1}.'.format(i, input.name))

    logger.debug('List outputs:')
    for i, output in enumerate(onnx_outputs):
        logger.debug('Output {0} -> {1}.'.format(i, output))

    logger.debug('Gathering weights to dictionary.')
    weights = {}
    for onnx_w in onnx_weights:
        try:
            if len(onnx_w.ListFields()) < 4:
                onnx_extracted_weights_name = onnx_w.ListFields()[1][1]
            else:
                onnx_extracted_weights_name = onnx_w.ListFields()[2][1]
            weights[onnx_extracted_weights_name] = numpy_helper.to_array(onnx_w)
        except:
            onnx_extracted_weights_name = onnx_w.ListFields()[3][1]
            weights[onnx_extracted_weights_name] = numpy_helper.to_array(onnx_w)

        logger.debug('Found weight {0} with shape {1}.'.format(
            onnx_extracted_weights_name,
            weights[onnx_extracted_weights_name].shape))

    layers = dict()
    lambda_funcs = dict()
    keras_outputs = []
    keras_inputs = []

    for i, input_name in enumerate(input_names):
        for onnx_i in onnx_inputs:
            if onnx_i.name == input_name:
                dtype = None if input_types is None else input_types[i]
                input_shape = [i.dim_value for i in onnx_i.type.tensor_type.shape.dim]
                input_shape = [shape if shape != 0 else None for shape in input_shape]
                if len(input_shape) <= 1:
                    input_tensor = keras.layers.InputLayer(input_shape=input_shape, name=input_name, dtype=dtype).output
                    layers[input_name] = input_tensor[0]
                    keras_inputs.append(input_tensor)

                else:
                    batch_size = input_shape[0]
                    input_shape = input_shape[1:]
                    if batch_size is None:
                        layers[input_name] = keras.layers.InputLayer(
                            input_shape=input_shape, name=input_name, dtype=dtype).output
                    else:
                        layers[input_name] = keras.layers.InputLayer(
                            input_shape=input_shape, name=input_name, dtype=dtype, batch_size=batch_size).output

                    keras_inputs.append(layers[input_name])

                logger.debug('Found input {0} with shape {1}'.format(input_name, input_shape))

    keras_middle_outputs = {}
    error_info = None
    try:
        # Convert every operation separable
        node_names = []
        embedding_weights_mapping = {}
        for node_index, node in enumerate(onnx_nodes):
            node_type = node.op_type
            node_params = onnx_node_attributes_to_dict(node.attribute)
            # Add global converter info:
            node_params['change_ordering'] = change_ordering
            node_params['name_policy'] = name_policy

            node_name = str(node.output[0])
            keras_names = []
            for output_index, output in enumerate(node.output):
                if name_policy == 'short':
                    keras_name = keras_name_i = str(output)[:8]
                    suffix = 1
                    while keras_name_i in node_names:
                        keras_name_i = keras_name + '_' + str(suffix)
                        suffix += 1
                    keras_names.append(keras_name_i)
                elif name_policy == 'renumerate':
                    postfix = node_index if len(node.output) == 1 else "%s_%s" % (node_index, output_index)
                    keras_names.append('LAYER_%s' % postfix)
                elif name_policy == 'attach_weights_name':
                    attached_weights_names = []
                    for node_input in node.input:
                        if node_input in weights:
                            weight_name = ".".join(node_input.split(".")[:-1])
                            attached_weights_names.append(weight_name)
                    set_weights_names = set(attached_weights_names)
                    set_weights_names = "__".join(set_weights_names)
                    layer_name = output.replace(":", "_")
                    while not (str.isalpha(layer_name[0]) or str.isdigit(layer_name[0]) or layer_name[0] == "."):
                        layer_name = layer_name[1:]

                    if layer_name == "":
                        layer_name = str(uuid.uuid4())[:10]

                    if set_weights_names:
                        layer_name = f"{layer_name}__{set_weights_names}"

                    keras_names.append(layer_name)
                else:
                    output = output.replace(":", "_")
                    keras_names.append(output)
            keras_names = [k.lstrip("/") for k in keras_names]
            if len(node.output) != 1:
                logger.warning('Trying to convert multi-output node')
                node_params['_outputs'] = list(node.output)
                node_names.extend(keras_names)
            else:
                keras_names = keras_names[0]
                node_names.append(keras_names)
            pattern = r'[#:@]'  # Example pattern to match #, /, and :
            cleaned_node_name = re.sub(pattern, '_', node.name.rstrip("/").lstrip("/"))
            if len(cleaned_node_name) == 0:
                cleaned_node_name = re.sub(pattern, '_', node_name.rstrip("/").lstrip("/"))
            node_params['cleaned_name'] = f'{cleaned_node_name}_tl'
            logger.debug('######')
            logger.debug(f"{node_index/len(onnx_nodes):.1%} completed")
            logger.debug('...')
            logger.debug('Converting ONNX operation')
            logger.debug('type: %s', node_type)
            logger.debug('node_name: %s', node_name)
            logger.debug('node_params: %s', node_params)
            logger.debug('...')

            logger.debug('Check if all inputs are available:')
            if len(node.input) == 0 and node_type != 'Constant':
                raise AttributeError('Operation doesn\'t have an input. Aborting.')
            for i, node_input in enumerate(node.input):
                logger.debug('Check input %i (name %s).', i, node_input)

                # for case of weights sharing, map the shared weights to determine
                # if a Gather layer is an embedding layer
                if node_type == 'Identity' and node_input in weights:
                    embedding_weights_mapping[node_name] = node_input

                # check conditions for embedding layer
                is_in_weights = node_input in weights  # is this node input in weights
                is_mapped_to_weights = embedding_weights_mapping.get(node_input,
                                                                     '') in weights  # is this node inputs weights are shared with other input
                is_embedding = (
                                       is_in_weights or is_mapped_to_weights) and i == 0  # if either is true this layer is a possible embedding layer

                # if a layer is of type Gather and its input is in weights (or mapped to a weights input)
                # it's an embedding layer
                if node_type == "Gather" and is_embedding:
                    node_params['is_embedding'] = True

                if node_input not in layers:
                    logger.debug('The input not found in layers / model inputs.')
                    if node_input in weights:
                        logger.debug('Found in weights, add as a numpy constant.')
                        layers[node_input] = weights[node_input]
                    else:
                        if node_input == "" and node_type in ('Pad', 'Resize', 'Clip', 'LSTM', 'GRU'):
                            continue
                        else:
                            raise AttributeError('Current node is not in weights / model inputs / layers.')
            else:
                logger.debug('... found all, continue')

            keras.backend.set_image_data_format('channels_first')
            try:
                layer_converter_func = AVAILABLE_CONVERTERS[node_type]
            except KeyError:
                raise UnsupportedLayer(node_type)
            layer_converter_func(
                node,
                node_params,
                layers,
                lambda_funcs,
                node_name,
                keras_names
            )
            # remove node inputs
            for inp in node.input:
                keras_middle_outputs.pop(inp, None)
            # add node to middle map
            if tf.is_tensor(layers[node_name]):
                keras_middle_outputs[node_name] = layers[node_name]

            if isinstance(keras_names, list):
                keras_names = keras_names[0]

            try:
                logger.debug('Output TF Layer -> ' + str(layers[keras_names]))
            except KeyError:
                pass

        # Check for terminal nodes
        for layer in onnx_outputs:
            if layer in layers:
                keras_outputs.append(layers[layer])

        model = keras.models.Model(inputs=keras_inputs, outputs=keras_outputs)
    except Exception as e:
        if not allow_partial_compilation:
            raise

        if len(keras_middle_outputs) == 0:
            raise e

        error_info = repr(e)
        if isinstance(e, UnsupportedLayer):
            error_info = e.layer_description

        keras_outputs = list(keras_middle_outputs.values())
        try:
            model = keras.models.Model(inputs=keras_inputs, outputs=keras_outputs)
        except:
            raise e

    if change_ordering:
        change_ord_axes_map = {
            3: 2,
            1: 3,
            -1: 1
        }

        import numpy as np
        conf = model.get_config()

        for layer in conf['layers']:
            if layer['config'] and 'shared_axes' in layer['config']:
                # TODO: check axes first (if it's not 4D tensor)
                layer['config']['shared_axes'] = [1, 2]

            if layer['config'] and 'batch_input_shape' in layer['config']:
                layer['config']['batch_input_shape'] = \
                    tuple(np.reshape(np.array(
                        [
                            [None] +
                            list(layer['config']['batch_input_shape'][2:][:]) +
                            [layer['config']['batch_input_shape'][1]]
                        ]), -1
                    ))
            if layer['config'] and 'target_shape' in layer['config']:
                if len(list(layer['config']['target_shape'][1:][:])) > 0:
                    layer['config']['target_shape'] = \
                        tuple(np.reshape(np.array(
                            list(layer['config']['target_shape'][1:]) +
                            [layer['config']['target_shape'][0]]
                        ), -1), )

            if layer['config'] and 'data_format' in layer['config']:
                layer['config']['data_format'] = 'channels_last'
            if layer['config'] and 'axis' in layer['config']:
                axis = layer['config']['axis']
                # BatchNorm wrap axis with ListWrapper instead single INT value
                if isinstance(axis, (tuple, list)):
                    axis = axis[0]
                layer['config']['axis'] = change_ord_axes_map.get(axis, layer['config']['axis'])

        for layer in conf['layers']:
            if 'function' in layer['config'] and layer['config']['function'][1] is not None:
                kerasf = list(layer['config']['function'])
                dargs = list(kerasf[1])
                func = lambda_funcs.get(layer['name'])

                if func:
                    # ReduceSum operation has 'axis' param as array of ints. When onnx uses ReduceSum
                    # to reproduce SoftMax - dargs become something like [[1]] (list of lists)
                    # that why we handle collections.Iterable
                    if len(dargs) > 1 or isinstance(dargs[0], (tuple, list)):
                        params = inspect.signature(func).parameters
                        i = list(params.keys()).index('axes') if ('axes' in params) else -1

                        if i > 0:
                            i -= 1
                            axes = list(range(len(dargs[i].shape)))
                            axes = axes[0:1] + axes[2:] + axes[1:2]
                            dargs[i] = np.transpose(dargs[i], axes)

                        i = list(params.keys()).index('axis') if ('axis' in params) else -1

                        if i > 0:
                            i -= 1
                            axis = np.array(dargs[i])
                            axes_map = np.array([0, 3, 1, 2])
                            # to list because some tf operations check only for core python types (e.g tf.norm)
                            dargs[i] = axes_map[axis].tolist()
                    else:
                        # if map exits will change else will remain the same
                        dargs[0] = change_ord_axes_map.get(dargs[0], dargs[0])

                kerasf[1] = tuple(dargs)
                layer['config']['function'] = tuple(kerasf)

        keras.backend.set_image_data_format('channels_last')
        model_tf_ordering = keras.models.Model.from_config(conf, custom_objects=onnx_custom_objects_map)

        for dst_layer, src_layer, conf in zip(model_tf_ordering.layers, model.layers, conf['layers']):
            W = src_layer.get_weights()
            # TODO: check axes first (if it's not 4D tensor)
            if conf['config'] and 'shared_axes' in conf['config']:
                W[0] = W[0].transpose(1, 2, 0)
            dst_layer.set_weights(W)

        model = model_tf_ordering

    keras.backend.set_image_data_format(keras_fmt)

    response = ConvertedResponse(model, error_info)
    return response


def extract_op_node(node_graph, layers, lambda_funcs, keras_names, change_ordering, name_policy):
    op_node = None
    for node_i, node in enumerate(node_graph):
        if node.op_type == 'Constant':
            node_params = onnx_node_attributes_to_dict(node.attribute)
            # Add global converter info:
            node_params['change_ordering'] = change_ordering
            node_params['name_policy'] = name_policy
            node_name = str(node.output[0])

            AVAILABLE_CONVERTERS[node.op_type](
                node,
                node_params,
                layers,
                lambda_funcs,
                node_name,
                keras_names
            )
        else:  # op type
            if op_node is not None:
                raise NotImplementedError('Not Implemented: inner graph in If node with multiple operator nodes')
            op_node = node
    if op_node is None:
        raise NotImplementedError('Something is off with If node')
    return op_node



from typing import Any, Optional, List

from keras.layers import Layer, TFOpLambda
import tensorflow as tf
import numpy as np


# this custom layer needed because of a tensorflow bug on einsum serielization
class OnnxEinsumLayer(Layer):
    """

    Args:
        equation: str
        constant_input: Optional[List[float]]
        constant_place: Optional[int]
    """

    def __init__(self, equation: str, constant_input: Optional[List[float]], constant_place: Optional[int], **kwargs):
        super().__init__(**kwargs)
        self.equation = equation
        if constant_input is not None:
            if hasattr(constant_input, 'numpy'):
                constant_input = constant_input.numpy()
            if not isinstance(constant_input, np.ndarray):
                constant_input = np.array(constant_input)
            self.constant_input = constant_input
        self.constant_place = constant_place

    def call(self, inputs, *args, **kwargs):
        if self.constant_input is not None:
            if self.constant_place == 1:
                inputs = [inputs, self.constant_input]
            else:
                inputs = [self.constant_input, inputs]

        return tf.einsum(self.equation, *inputs)

    def get_config(self):
        config = super().get_config()
        config.update({
            "equation": self.equation,
            "constant_input": self.constant_input,
            "constant_place": self.constant_place
        })
        return config



from keras.layers import Layer
import tensorflow as tf


class OnnxHardSigmoid(Layer):
    def __init__(self, alpha: float = 0.2, beta: float = 0.5, **kwargs):
        super().__init__(**kwargs)
        self.alpha = alpha
        self.beta = beta

    def call(self, inputs, **kwargs):
        x = tf.multiply(inputs, self.alpha)
        x = tf.add(x, self.beta)
        x = tf.clip_by_value(x, 0., 1.)
        return x

    def get_config(self):
        config = super().get_config()
        config.update({
            "alpha": self.alpha,
            "beta": self.beta,
        })
        return config



import tensorflow as tf
from .tfops_funcs import tf_signal_irfft, tf_signal_ifft, tf_signal_rfft, tf_signal_fft


def convert_dft(node, params, layers, lambda_func, node_name, keras_name):
    # Currently, there's no way to run this - pytorch export not supported + Onnx Runtime version is too advanced.
    raise AttributeError("DFT could not yet be converted - contact Tensorleap for support")
    axis = params.get('axis', 1)
    inverse = bool(params.get('inverse', 0))
    onesided = bool(params.get('onesided', 0))
    input_0 = layers[node.input[0]]
    if len(node.input[0]) == 2:
        fft_length = layers[node.input[0]]
    else:
        fft_length = None
    rank_tensor = len(input_0.shape)
    rotated_axis = False
    if axis != rank_tensor-1 and axis != -1: # tf.fft works only on last dimension - permuting
        if axis == 0:
            raise AttributeError("FFT on the batch dimension isn't convertable")
        output = tf.keras.layers.Permute((axis, rank_tensor-1), name=f"{params['cleaned_name']}_dft_permute_1")(input_0)
        rotated_axis = True
    else:
        output = input_0
    if inverse and onesided:
        output = tf_signal_irfft(input_0, fft_length=fft_length, tf_name=f"{params['cleaned_name']}_dft_irfft")
    if inverse and not onesided:
        output = tf_signal_ifft(input_0, fft_length=fft_length, tf_name=f"{params['cleaned_name']}_dft_ifft")
    if onesided:
        output = tf_signal_rfft(input_0, fft_length=fft_length, tf_name=f"{params['cleaned_name']}_dft_rfft")
    if not onesided:
        output = tf_signal_fft(input_0, fft_length=fft_length, tf_name=f"{params['cleaned_name']}_dft_fft")
    if rotated_axis:
        layers[node_name] = tf.keras.layers.Permute((rank_tensor-1, axis),
                                                    name=f"{params['cleaned_name']}_dft_permute_2")(output)
    else:
        layers[node_name] = output



from onnx2kerastl.customonnxlayer.onnxeinsum import OnnxEinsumLayer
from onnx2kerastl.customonnxlayer.onnxlstm import OnnxLSTM

onnx_custom_objects_map = {
    "OnnxLSTM": OnnxLSTM,
}

onnx_custom_layers = {
    "OnnxEinsumLayer": OnnxEinsumLayer
}



import argparse
import numpy as np
import onnx
from onnx2kerastl import onnx_to_keras
import tensorflow as tf
from keras_data_format_converter import convert_channels_first_to_last

def convert_onnx_to_keras(onnx_model_path, transform_io:bool = True):
    # Load ONNX model
    save_model_path = onnx_model_path.replace('.onnx', '.h5')
    onnx_model = onnx.load(onnx_model_path)
    
    # Extract input feature names from the model
    input_features = [inp.name for inp in onnx_model.graph.input]
    
    # Convert ONNX model to Keras
    keras_model = onnx_to_keras(onnx_model, input_names=input_features,
                                name_policy='attach_weights_name', allow_partial_compilation=False).converted_model

    # Convert from channels-first to channels-last format
    final_model = convert_channels_first_to_last(keras_model, should_transform_inputs_and_outputs=transform_io,
                                                 verbose=True)

    # Save the final Keras model
    final_model.save(save_model_path)
    print(f"Model saved to {save_model_path}")


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Convert ONNX model to Keras')
    parser.add_argument('onnx_path', type=str, help='Path to the input ONNX model')    
    parser.add_argument('transform_input_output', type=bool, help='Whether to transform input and output data format')
    args = parser.parse_args()
    
    # Convert input_shape string to tuple of integers
    
    convert_onnx_to_keras(args.onnx_path, args.transform_input_output)


import numpy as np
import tensorflow as tf
from .utils import is_numpy
from .tfops_funcs import tf_cast, tf_one_hot
from keras import backend as K

def convert_constant(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert Constant layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    layers[node_name] = params['value']


def convert_constant_of_shape(node, params, layers, lambda_func, node_name, keras_name):
    value = params.get('value')
    if value is None:
        raise NotImplementedError("ConstantOfShape should have a value param")

    input_0 = layers[node.input[0]]

    if not is_numpy(input_0) and not isinstance(input_0, list) and K.is_keras_tensor(input_0):
        # Boolean case
        if value.dtype == np.bool_:
            layers[node_name] = tf.fill(layers[node.input[0]], tf.constant(value.item(), dtype=tf.bool))
        else:
            # Non-boolean case
            layers[node_name] = tf.ones(layers[node.input[0]], dtype=tf.as_dtype(value.dtype)) * value
    else:
        # Handle numpy inputs or non-Keras tensors
        if value.dtype == np.bool_:
            layers[node_name] = np.full(layers[node.input[0]], value.item(), dtype=bool)
        else:
            layers[node_name] = np.ones(layers[node.input[0]], dtype=value.dtype) * value



def convert_one_hot(node, params, layers, lambda_func, node_name, keras_name):
    axis = params.get('axis', -1)
    layers[node_name] = tf_one_hot(indices=tf_cast(layers[node.input[0]],
                                                   tf.int64,
                                                   tf_name=f"{params['cleaned_name']}_onehot_cast"),
                                   depth=int(layers[node.input[1]]),
                                   off_value=layers[node.input[2]][0],
                                   on_value=layers[node.input[2]][1],
                                   axis=axis,
                                   tf_name=f"{params['cleaned_name']}_onehot"
                                   )




import numpy as np
import keras
import logging

from .utils import is_numpy, ensure_tf_type
from .tfops_funcs import tf_tensor_scatter_nd_update, tf_maximum, tf_minimum, tf_cast, tf_expand_dims, tf_repeat,\
    tf_equal, tf_where, tf_round, tf_sign, tf_abs, tf_math_mod, tf_bitwise_left_shift, tf_bitwise_right_shift,\
    tf_logical_not, tf_add
import tensorflow as tf
from tensorflow.python.framework.ops import EagerTensor


def _is_integer_type(dtype) -> bool:
    return dtype in (tf.int32, tf.int64, tf.int16, tf.int8, np.int32, np.int64, np.int16, np.int8)


def convert_elementwise_div(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert element-wise division
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    logger = logging.getLogger('onnx2keras.div')

    if len(node.input) != 2:
        raise AttributeError('Number of inputs is not equal 2 for element-wise layer')

    input_0 = layers[node.input[0]]
    input_1 = layers[node.input[1]]

    try:
        logger.debug('Divide numpy arrays.')
        div = input_0 / input_1
        if _is_integer_type(input_0.dtype) and _is_integer_type(input_1.dtype):
            div = tf_cast(div, input_0.dtype, tf_name=f"{params['cleaned_name']}_div_cast")
        if hasattr(div, 'numpy'):
            div = div.numpy()
        layers[node_name] = div

    except (IndexError, ValueError):
        logger.debug('Convert inputs to Keras/TF layers if needed.')

        def target_layer(x):
            import tensorflow as tf
            layer = tf.divide(
                x[0],
                x[1]
            )
            return layer

        lambda_layer = keras.layers.Lambda(target_layer, name=f"{params['cleaned_name']}_div")
        layers[node_name] = lambda_layer([input_0, input_1])
        lambda_func[keras_name] = target_layer


def convert_elementwise_add(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert element-wise add.
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    logger = logging.getLogger('onnx2keras.add')

    if len(node.input) != 2:
        raise AttributeError('Number of inputs is not equal to 2 for element-wise layer')

    input_0 = layers[node.input[0]]
    input_1 = layers[node.input[1]]

    input_0_is_constant = is_numpy(input_0) or isinstance(input_0, EagerTensor)
    input_1_is_constant = is_numpy(input_1) or isinstance(input_1, EagerTensor)

    try:
        if not input_0_is_constant and not input_1_is_constant:
            # Both inputs are variables
            if len(input_0.shape) != len(input_1.shape):
                # Use TensorFlow add to handle shape differences
                layers[node_name] = tf_add(input_0, input_1, tf_name=f"{params['cleaned_name']}_add")
            else:
                # Use Keras Add layer
                layers[node_name] = keras.layers.Add(name=f"{params['cleaned_name']}_add")([input_0, input_1])
        else:
            raise ValueError('Operands are different.')
    except (IndexError, ValueError):
        logger.warning('Failed to use keras.layers.Add. Fallback to Lambda layer.')

        if input_0_is_constant and not input_1_is_constant:
            # input_0 is constant, input_1 is variable
            constant_value = np.asarray(tf.cast(input_0, dtype=input_1.dtype))
            variable_input = input_1

            if np.all(constant_value == constant_value.flat[0]):
                # Constant tensor has the same value throughout
                const_val = constant_value.flat[0]
                layers[node_name] = keras.layers.Lambda(
                    lambda x: x + const_val,
                    name=params['cleaned_name']
                )(variable_input)
            else:
                # Embedding the constant tensor
                layers[node_name] = keras.layers.Lambda(
                    lambda x: x + constant_value,
                    name=params['cleaned_name']
                )(variable_input)

        elif not input_0_is_constant and input_1_is_constant:
            # input_0 is variable, input_1 is constant
            constant_value = np.asarray(tf.cast(input_1, dtype=input_0.dtype))
            variable_input = input_0

            if np.all(constant_value == constant_value.flat[0]):
                # Constant tensor has the same value throughout
                const_val = constant_value.flat[0]
                layers[node_name] = keras.layers.Lambda(
                    lambda x: x + const_val,
                    name=params['cleaned_name']
                )(variable_input)
            else:
                # Embedding the constant tensor
                layers[node_name] = keras.layers.Lambda(
                    lambda x: x + constant_value,
                    name=params['cleaned_name']
                )(variable_input)
        else:
            # Both inputs are constants; compute the result now
            layers[node_name] = input_0 + input_1



def convert_elementwise_mul(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert element-wise mul.
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    logger = logging.getLogger('onnx2keras.mul')

    if len(node.input) != 2:
        raise AttributeError('Number of inputs is not equal to 2 for element-wise layer')

    input_0 = layers[node.input[0]]
    input_1 = layers[node.input[1]]

    input_0_is_constant = is_numpy(input_0) or isinstance(input_0, EagerTensor)
    input_1_is_constant = is_numpy(input_1) or isinstance(input_1, EagerTensor)

    try:
        if not input_0_is_constant and not input_1_is_constant:
            mul = keras.layers.Multiply(name=f"{params['cleaned_name']}_mul")
            layers[node_name] = mul([input_0, input_1])
        else:
            raise ValueError('Operands are different.')

    except (IndexError, ValueError):
        logger.warning('Failed to use keras.layers.Multiply. Fallback to Lambda layer.')

        if input_0_is_constant and not input_1_is_constant:
            # input_0 is constant, input_1 is variable
            constant_value = np.asarray(tf.cast(input_0, dtype=input_1.dtype))
            variable_input = input_1

            if np.all(constant_value == constant_value.flat[0]):
                # Constant tensor has the same value throughout
                const_val = constant_value.flat[0]
                layers[node_name] = keras.layers.Lambda(
                    lambda x: x * const_val,
                    name=params['cleaned_name']
                )(variable_input)
            else:
                # Cannot avoid embedding the constant tensor
                layers[node_name] = keras.layers.Lambda(
                    lambda x: x * constant_value,
                    name=params['cleaned_name']
                )(variable_input)

        elif not input_0_is_constant and input_1_is_constant:
            # input_0 is variable, input_1 is constant
            constant_value = np.asarray(tf.cast(input_1, dtype=input_0.dtype))
            variable_input = input_0

            if np.all(constant_value == constant_value.flat[0]):
                # Constant tensor has the same value throughout
                const_val = constant_value.flat[0]
                layers[node_name] = keras.layers.Lambda(
                    lambda x: x * const_val,
                    name=params['cleaned_name']
                )(variable_input)
            else:
                # Cannot avoid embedding the constant tensor
                layers[node_name] = keras.layers.Lambda(
                    lambda x: x * constant_value,
                    name=params['cleaned_name']
                )(variable_input)
        else:
            # Both inputs are constants; compute the result now
            layers[node_name] = input_0 * input_1


def convert_elementwise_sub(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert element-wise sub.
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    logger = logging.getLogger('onnx2keras.sub')

    if len(node.input) != 2:
        raise AttributeError('Number of inputs is not equal 2 for element-wise layer')

    input_0 = layers[node.input[0]]
    input_1 = layers[node.input[1]]
    
    input_0_is_constant = is_numpy(input_0) or isinstance(input_0, EagerTensor)
    input_1_is_constant = is_numpy(input_1) or isinstance(input_1, EagerTensor)

    try:
        if not input_0_is_constant and not input_1_is_constant:
            sub = keras.layers.Subtract(name=f"{params['cleaned_name']}_sub")
            layers[node_name] = sub([input_0, input_1])
        else:
            raise ValueError('Operands are different.')

    except (IndexError, ValueError):
        logger.warning('Failed to use keras.layers.Subtract. Fallback to Lambda layer.')

        if input_0_is_constant and not input_1_is_constant:
            # input_0 is constant, input_1 is variable: constant - variable
            constant_value = np.asarray(tf.cast(input_0, dtype=input_1.dtype))
            variable_input = input_1

            if np.all(constant_value == constant_value.flat[0]):
                # Constant tensor has the same value throughout
                const_val = constant_value.flat[0]
                layers[node_name] = keras.layers.Lambda(
                    lambda x: const_val - x,
                    name=params['cleaned_name']
                )(variable_input)
            else:
                # Cannot avoid embedding the constant tensor
                layers[node_name] = keras.layers.Lambda(
                    lambda x: constant_value - x,
                    name=params['cleaned_name']
                )(variable_input)

        elif not input_0_is_constant and input_1_is_constant:
            # input_0 is variable, input_1 is constant: variable - constant
            constant_value = np.asarray(tf.cast(input_1, dtype=input_0.dtype))
            variable_input = input_0

            if np.all(constant_value == constant_value.flat[0]):
                # Constant tensor has the same value throughout
                const_val = constant_value.flat[0]
                layers[node_name] = keras.layers.Lambda(
                    lambda x: x - const_val,
                    name=params['cleaned_name']
                )(variable_input)
            else:
                # Cannot avoid embedding the constant tensor
                layers[node_name] = keras.layers.Lambda(
                    lambda x: x - constant_value,
                    name=params['cleaned_name']
                )(variable_input)
        else:
            # Both inputs are constants; compute the result now
            layers[node_name] = input_0 - input_1



def convert_min(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert Min layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    if len(node.input) < 2:
        raise AttributeError('Less than 2 inputs for min layer.')

    inputs = [
        ensure_tf_type(layers[inp], name="%s_const%i" % (keras_name, i + 1))
        for i, inp in enumerate(node.input)
    ]

    # Broadcast the inputs to the same shape
    input1, input2 = inputs
    # Applying the minimum operation
    min_output = tf_minimum(input1, input2, tf_name=f"{params['cleaned_name']}_min")

    layers[node_name] = min_output


def convert_max(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert Max layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    if len(node.input) < 2:
        raise AttributeError('Less than 2 inputs for max layer.')

    inputs = [
        ensure_tf_type(layers[inp], name="%s_const%i" % (keras_name, i + 1))
        for i, inp in enumerate(node.input)
    ]

    # Broadcast the inputs to the same shape
    input1, input2 = inputs
    # Applying the maximum operation
    max_output = tf_maximum(input1, input2, tf_name=f"{params['cleaned_name']}_maximum")

    layers[node_name] = max_output


def convert_mean(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert Mean layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    :TODO: Test if this supports multidirectional (i.e., Numpy-style) broadcasting as required
    """
    if len(node.input) < 2:
        assert AttributeError('Less than 2 inputs for mean layer.')

    inputs = list()
    for i, inp in enumerate(node.input):
        input_ = ensure_tf_type(layers[inp], layers[list(layers)[0]], name="%s_const%i" % (keras_name, i + 1))
        inputs.append(input_)
    layers[node_name] = keras.layers.Average(name=f"{params['cleaned_name']}_mean")(inputs)


def convert_equal(node, params, layers, lambda_func, node_name, keras_name):
    layers[node_name] = tf_equal(layers[node.input[0]], layers[node.input[1]],
                                 tf_name=f"{params['cleaned_name']}_equal")


def convert_where(node, params, layers, lambda_func, node_name, keras_name):
    if layers[node.input[0]].dtype != tf.bool:
        casted = tf_cast(layers[node.input[0]], tf.bool, tf_name=f"{params['cleaned_name']}_cast")
    else:
        casted = layers[node.input[0]]
    if layers[node.input[1]].dtype == np.int64 and is_numpy(layers[node.input[1]]):
        # serialization doesn't work well for first argument if it is np array of type int64
        layers[node_name] = tf_where(tf_logical_not(casted,
                                                    tf_name=f"{params['cleaned_name']}_not"
                                                    ),
                                     layers[node.input[2]],
                                     layers[node.input[1]],
                                     tf_name=f"{params['cleaned_name']}_where_1")
    else:
        try:
            layers[node_name] = tf_where(casted, layers[node.input[1]], layers[node.input[2]],
                                         tf_name=f"{params['cleaned_name']}_where_2")
        except Exception as e:
            print(1)


def convert_scatter_nd(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert ScatterND layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    :TODO: Test if this supports multidirectional (i.e., Numpy-style) broadcasting as required
    """
    if len(node.input) < 3:
        assert AttributeError('Less than 3 inputs')

    data = ensure_tf_type(layers[node.input[0]])
    indices = ensure_tf_type(layers[node.input[1]])
    updates = ensure_tf_type(layers[node.input[2]])
    layers[node_name] = tf_tensor_scatter_nd_update(data, indices, updates,
                                                    tf_name=f"{params['cleaned_name']}_scatter_nd")


def convert_round(node, params, layers, lambda_func, node_name, keras_name):
    layers[node_name] = tf_round(layers[node.input[0]], tf_name=f"{params['cleaned_name']}_round")


def convert_mod(node, params, layers, lambda_func, node_name, keras_name):
    input_0 = layers[node.input[0]]
    input_1 = layers[node.input[1]]
    if params.get('fmod') == 1:
        sign = tf_sign(layers[node.input[0]], tf_name=f"{params['cleaned_name']}_mod_sign")
        input_0 = tf_abs(layers[node.input[0]], tf_name=f"{params['cleaned_name']}_abs_0")
        input_1 = tf_abs(layers[node.input[1]], tf_name=f"{params['cleaned_name']}_abs_1")
        layers[node_name] = tf_math_mod(input_0, input_1, tf_name=f"{params['cleaned_name']}_mod") * sign
    else:
        layers[node_name] = tf_math_mod(input_0, input_1, tf_name=f"{params['cleaned_name']}_mod")


def convert_bitshift(node, params, layers, lambda_func, node_name, keras_name):
    direction = params.get("direction").decode()
    if direction == "LEFT":
        shifter_pointer = tf_bitwise_left_shift
    elif direction == "RIGHT":
        shifter_pointer = tf_bitwise_right_shift
    else:
        raise AttributeError("Onnx2Kerras cannot convert the BitShift operator"
                             " since the 'direction' attribute was missing")
    layers[node_name] = shifter_pointer(tf_cast(layers[node.input[0]], tf.uint64,
                                                tf_name=f"{params['cleaned_name']}_bitshift_cast_0"),
                                        tf_cast(layers[node.input[1]], tf.uint64,
                                                tf_name=f"{params['cleaned_name']}_bitshift_cast_1"),
                                        tf_name=f"{params['cleaned_name']}_bitshift")



import numpy as np
import tensorflow as tf
from typing import Callable
from keras import backend as K
import logging
logger = logging.getLogger('onnx2keras.tfops_funcs')

layer_names_counter = {}


def named_tfop(func: Callable):
    def wrapped_function(*args, tf_name=None, **kwargs):
        result = func(*args, **kwargs)
        if tf_name is None or tf_name=="" or not isinstance(tf_name, str):
            raise ValueError(f"The layer {result.node.layer} with name"
                             f" {result.node.layer.name} was provided with an empty or None Name")
        if not isinstance(result, (tf.Tensor, np.ndarray)):
            if tf_name not in layer_names_counter:
                layer_names_counter[tf_name] = 0
            else:
                layer_names_counter[tf_name] = layer_names_counter[tf_name] + 1
                tf_name = tf_name + f"_{layer_names_counter[tf_name]}"
                if not isinstance(result, (tf.Tensor, np.ndarray)):
                    logger.debug(f"The op {result.node.layer.symbol} with name"
                                     f"{result.node.layer.name} has a duplicate name {tf_name}")
            result.node.layer._name = tf_name
        return result
    return wrapped_function


tf_cast = named_tfop(tf.cast)
tf_shape = named_tfop(tf.shape)
tf_math_abs = named_tfop(tf.math.abs)
tf_reshape = named_tfop(tf.reshape)
tf_stack = named_tfop(tf.stack)
tf_add = named_tfop(tf.add)
tf_image_resize = named_tfop(tf.image.resize)
tf_multiply = named_tfop(tf.multiply)
tf_clip_by_value = named_tfop(tf.clip_by_value)
tf_math_negative = named_tfop(tf.math.negative)
tf_tensor_scatter_nd_update = named_tfop(tf.tensor_scatter_nd_update)
K_mean = named_tfop(K.mean)
tf_math_reduce_prod = named_tfop(tf.math.reduce_prod)
tf_math_reduce_min = named_tfop(tf.math.reduce_min)
tf_math_pow = named_tfop(tf.math.pow)
tf_math_sqrt = named_tfop(tf.math.sqrt)
tf_strided_slice = named_tfop(tf.strided_slice)
tf_squeeze = named_tfop(tf.squeeze)
tf_argmax = named_tfop(tf.argmax)
tf_expand_dims = named_tfop(tf.expand_dims)
tf_maximum = named_tfop(tf.maximum)
tf_minimum = named_tfop(tf.minimum)
tf_repeat = named_tfop(tf.repeat)
tf_matmul = named_tfop(tf.matmul)
tf_concat = named_tfop(tf.concat)
tf_transpose = named_tfop(tf.transpose)
tf_math_reduce_variance = named_tfop(tf.math.reduce_variance)
tf_math_reduce_mean = named_tfop(tf.math.reduce_mean)
tf_sqrt = named_tfop(tf.sqrt)
tf_where = named_tfop(tf.where)
tf_gather = named_tfop(tf.gather)
tf_range = named_tfop(tf.range)
tf_abs = named_tfop(tf.abs)
tf_reduce_sum = named_tfop(tf.reduce_sum)
tf_pad = named_tfop(tf.pad)
tf_math_erf = named_tfop(tf.math.erf)
tf_math_reciprocal = named_tfop(tf.math.reciprocal)
tf_logical_not = named_tfop(tf.logical_not)
tf_equal = named_tfop(tf.equal)
tf_tile = named_tfop(tf.tile)
tf_math_minimum = named_tfop(tf.math.minimum)
tf_math_maximum = named_tfop(tf.math.maximum)
tf_math_sign = named_tfop(tf.math.sign)
tf_math_sin = named_tfop(tf.math.sin)
tf_math_cosh = named_tfop(tf.math.cosh)
tf_math_ceil = named_tfop(tf.math.ceil)
tf_math_acosh = named_tfop(tf.math.acosh)
tf_math_acos = named_tfop(tf.math.acos)
tf_math_asinh = named_tfop(tf.math.asinh)
tf_math_asin = named_tfop(tf.math.asin)
tf_math_atanh = named_tfop(tf.math.atanh)
tf_math_tan = named_tfop(tf.math.tan)
tf_math_atan = named_tfop(tf.math.atan)
tf_math_sinh = named_tfop(tf.math.sinh)
tf_math_less_equal = named_tfop(tf.math.less_equal)
tf_bitwise_invert = named_tfop(tf.bitwise.invert)
tf_bitwise_bitwise_and = named_tfop(tf.bitwise.bitwise_and)
tf_bitwise_bitwise_or = named_tfop(tf.bitwise.bitwise_or)
tf_bitwise_bitwise_xor = named_tfop(tf.bitwise.bitwise_xor)
tf_cos = named_tfop(tf.cos)
tf_math_greater = named_tfop(tf.math.greater)
tf_math_greater_equal = named_tfop(tf.math.greater_equal)
tf_logical_and = named_tfop(tf.logical_and)
tf_math_logical_xor = named_tfop(tf.math.logical_xor)
tf_math_logical_or = named_tfop(tf.math.logical_or)
tf_argmin = named_tfop(tf.argmin)
tf_one_hot = named_tfop(tf.one_hot)
tf_round = named_tfop(tf.round)
tf_math_cumsum = named_tfop(tf.math.cumsum)
tf_math_is_inf = named_tfop(tf.math.is_inf)
tf_math_is_nan = named_tfop(tf.math.is_nan)
tf_size = named_tfop(tf.size)
tf_linalg_det = named_tfop(tf.linalg.det)
tf_not_equal = named_tfop(tf.not_equal)
tf_gather_nd = named_tfop(tf.gather_nd)
tf_math_softplus = named_tfop(tf.math.softplus)
tf_math_tanh = named_tfop(tf.math.tanh)
tf_signal_irfft = named_tfop(tf.signal.irfft)
tf_signal_ifft = named_tfop(tf.signal.ifft)
tf_signal_rfft = named_tfop(tf.signal.rfft)
tf_signal_fft = named_tfop(tf.signal.fft)
tf_sign = named_tfop(tf.sign)
tf_abs = named_tfop(tf.abs)
tf_math_mod = named_tfop(tf.math.mod)
tf_bitwise_left_shift = named_tfop(tf.bitwise.left_shift)
tf_bitwise_right_shift = named_tfop(tf.bitwise.right_shift)
tf_rank = named_tfop(tf.rank)
tf_fill = named_tfop(tf.fill)
tf_image_non_max_suppression = named_tfop(tf.image.non_max_suppression)
tf_ones_like = named_tfop(tf.ones_like)
tf_image_crop_and_resize = named_tfop(tf.image.crop_and_resize)
tf_ones = named_tfop(tf.ones)
tf_math_floor = named_tfop(tf.math.floor)
tf_zeros_like = named_tfop(tf.zeros_like)
tf_tensor_scatter_nd_update = named_tfop(tf.tensor_scatter_nd_update)
tf_nn_avg_pool = named_tfop(tf.nn.avg_pool)
tf_nn_max_pool = named_tfop(tf.nn.max_pool)
tf_linalg_matmul = named_tfop(tf.linalg.matmul)


import keras
import numpy as np
import logging


def convert_upsample(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert upsample.
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    logger = logging.getLogger('onnx2keras.upsample')
    logger.warning('!!! EXPERIMENTAL SUPPORT (upsample) !!!')

    if "scales" in params:
        # for opset version - 7
        if len(node.input) != 1:
            raise AttributeError('Unsupported number of inputs')
        scale = np.uint8(params['scales'][-2:])
    else:
        # for opset version - 9+
        # Upsample since opset version 9 uses input[1] as 'scales' instead of attributes.
        scale = np.uint8(layers[node.input[1]][-2:])

    interpolation_mode = params['mode'].decode('utf-8')
    if interpolation_mode == 'nearest':
        interpolation = "nearest"
    elif interpolation_mode in ['bilinear', 'linear']:
        interpolation = "bilinear"
    elif interpolation_mode in "cubic":
        interpolation = "bicubic"
    else:
        logger.error(f'Cannot convert upsampling. interpolation mode: {interpolation_mode} is not supported')
        raise AssertionError(f'Cannot convert upsampling. interpolation mode: {interpolation_mode} is not supported')

    upsampling = keras.layers.UpSampling2D(size=scale, name=f"{params['cleaned_name']}_upsample_2d",
                                           interpolation=interpolation)

    layers[node_name] = upsampling(layers[node.input[0]])



import keras
import tensorflow as tf
from .tfops_funcs import tf_shape, tf_expand_dims, tf_cast, tf_reshape,\
    tf_math_minimum, tf_math_maximum, tf_range, tf_gather, tf_size, tf_math_floor, tf_concat


def convert_range(node, params, layers, lambda_func, node_name, keras_name):
    start_range = layers[node.input[0]]
    limit_range = layers[node.input[1]]
    delta_range = layers[node.input[2]]
    layers[node_name] = tf_range(start_range, limit_range, delta_range, tf_name=f"{params['cleaned_name']}_range")


def convert_gridsample(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert gridsample.
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    assert params['mode'].decode('ascii') == 'bilinear'
    assert params['padding_mode'].decode('ascii') == 'zeros'
    params['mode'] = params['mode'].decode('ascii')
    params['padding_mode'] = params['padding_mode'].decode('ascii')
    img = layers[node.input[0]]
    sample_grid = layers[node.input[1]]
    torch_shape = tf_shape(img, tf_name=f"{params['cleaned_name']}_gridsample_img_shape")
    max_xy = tf_expand_dims(
        tf_expand_dims(tf_expand_dims(tf.convert_to_tensor([torch_shape[3] - 1, torch_shape[2] - 1]),
                                      0,
                                      tf_name=f"{params['cleaned_name']}_gridsample_max_xy_expand_1"),
                       0,
                       tf_name=f"{params['cleaned_name']}_gridsample_max_xy_expand_2"),
        0,
        tf_name=f"{params['cleaned_name']}_gridsample_max_xy_expand_3")
    max_xy = tf_cast(max_xy, tf.float32, tf_name=f"{params['cleaned_name']}_gridsample_cast")

    if params['align_corners'] == 1:
        # Case when align_corners is 1
        grid_index_coords = 0.5 * (sample_grid + 1.) * max_xy
        grid_index_coords = grid_index_coords + 1
    else:
        # Case when align_corners is 0
        grid_index_coords = ((sample_grid + 1.) * max_xy - 1) / 2

    orig_query_shape = tf_shape(grid_index_coords,
                                tf_name=f"{params['cleaned_name']}_gridsample_coords_shape")
    query_points = tf_reshape(grid_index_coords, [orig_query_shape[0], -1, 2],
                              tf_name=f"{params['cleaned_name']}_gridsample_coords_reshape")
    padded_img = tf.keras.layers.ZeroPadding2D(padding=(1, 1), data_format="channels_first",
                                               name=f"{params['cleaned_name']}_gridsample_pad_img")(img)
    grid = tf.keras.layers.Permute((2, 3, 1),
                                   name=f"{params['cleaned_name']}_gridsample_rotate")(padded_img)
    indexing = 'ji'
    grid_shape = tf_shape(grid, tf_name=f"{params['cleaned_name']}_gridsample_grid_shape")
    query_shape = tf_shape(query_points, tf_name=f"{params['cleaned_name']}_query_shape")
    batch_size, height, width, channels = (
        grid_shape[0],
        grid_shape[1],
        grid_shape[2],
        grid_shape[3],
    )
    num_queries = query_shape[1]

    query_type = query_points.dtype
    grid_type = grid.dtype

    alphas = []
    floors = []
    ceils = []
    index_order = [0, 1] if indexing == "ij" else [1, 0]
    # unstacked_query_points = tf.unstack(query_points, axis=2, num=2)

    for i, dim in enumerate(index_order):
        queries = query_points[:, :, dim]
        # queries = unstacked_query_points[dim]

        size_in_indexing_dimension = grid_shape[i + 1]

        # max_floor is size_in_indexing_dimension - 2 so that max_floor + 1
        # is still a valid index into the grid.
        max_floor = tf_cast(size_in_indexing_dimension - 2, query_type,
                            tf_name=f"{params['cleaned_name']}_gridsample_cast_max")
        min_floor = tf.constant(0.0, dtype=query_type)
        floor = tf_math_minimum(
            tf_math_maximum(min_floor,
                            tf_math_floor(queries, tf_name=f"{params['cleaned_name']}_gridsample_floor"),
                            tf_name=f"{params['cleaned_name']}_gridsample_max"),
            max_floor,
            tf_name=f"{params['cleaned_name']}_gridsample_min"
        )
        int_floor = tf_cast(floor, tf.dtypes.int32, tf_name=f"{params['cleaned_name']}_gridsample_cast_floor")
        floors.append(int_floor)
        ceil = int_floor + 1
        ceils.append(ceil)

        # alpha has the same type as the grid, as we will directly use alpha
        # when taking linear combinations of pixel values from the image.
        alpha = tf_cast(queries - floor, grid_type, tf_name=f"{params['cleaned_name']}_gridsample_alpha")
        min_alpha = tf.constant(0.0, dtype=grid_type)
        max_alpha = tf.constant(1.0, dtype=grid_type)
        alpha = tf_math_minimum(tf_math_maximum(min_alpha,
                                                alpha,
                                                tf_name=f"{params['cleaned_name']}_gridsample_alpha_max"),
                                max_alpha,
                                tf_name=f"{params['cleaned_name']}_gridsample_alpha_min")

        # Expand alpha to [b, n, 1] so we can use broadcasting
        # (since the alpha values don't depend on the channel).
        alpha = tf_expand_dims(alpha, 2, tf_name=f"{params['cleaned_name']}_gridsample_expand_dim")
        alphas.append(alpha)

        flattened_grid = tf_reshape(grid, [batch_size * height * width, channels],
                                    tf_name=f"{params['cleaned_name']}_gridsample_flatenned_reshape")
        batch_offsets = tf_reshape(
            tf_range(batch_size, tf_name=f"{params['cleaned_name']}_gridsample_range") * height * width, [batch_size, 1],
            tf_name=f"{params['cleaned_name']}_gridsample_flatenned_offset"
        )

    # This wraps tf.gather. We reshape the image data such that the
    # batch, y, and x coordinates are pulled into the first dimension.
    # Then we gather. Finally, we reshape the output back. It's possible this
    # code would be made simpler by using tf.gather_nd.
    def gather(y_coords, x_coords, name=None):
        linear_coordinates = batch_offsets + y_coords * width + x_coords
        gathered_values = tf_gather(flattened_grid, linear_coordinates,
                                    tf_name=f"{params['cleaned_name']}_gridsample_gather_{name}")
        return tf_reshape(gathered_values, [batch_size, num_queries, channels],
                          tf_name=f"{params['cleaned_name']}_gridsample_reshape_{name}")

    # grab the pixel values in the 4 corners around each query point
    top_left = gather(floors[0], floors[1], "top_left")
    top_right = gather(floors[0], ceils[1], "top_right")
    bottom_left = gather(ceils[0], floors[1], "bottom_left")
    bottom_right = gather(ceils[0], ceils[1], "bottom_right")

    interp_top = alphas[1] * (top_right - top_left) + top_left
    interp_bottom = alphas[1] * (bottom_right - bottom_left) + bottom_left
    interp = alphas[0] * (interp_bottom - interp_top) + interp_top
    tf_reshaped_results = tf_reshape(interp,
                                     tf_concat([orig_query_shape[:-1], torch_shape[1:2]], axis=0,
                                               tf_name=f"{params['cleaned_name']}_gridsample_concat"),
                                     tf_name=f"{params['cleaned_name']}_gridsample_reshape_res")
    ret = tf.keras.layers.Permute((3, 1, 2),
                                  name=f"{params['cleaned_name']}_gridsample_last_permute")(tf_reshaped_results)
    layers[node_name] = ret


def convert_unique(node, params, layers, lambda_func, node_name, keras_name):
    to_sort = params.get('sorted', 1) == 1
    axis = params.get('axis')
    if axis is not None:
        raise AttributeError("Onnx2kerras: Unique does not currently support an operation on a non-flattened array")
    lambda_input = layers[node.input[0]]
    rev_idx_length = tf_size(lambda_input, tf_name=f"{params['cleaned_name']}_unique_size_1")

    def target_layer(x):
        input_keras = x
        if axis is None:
            input_final = tf.reshape(input_keras, [-1])
        res, rev_idx, count = tf.unique_with_counts(input_final)
        idx = tf.math.unsorted_segment_min(tf.range(tf.shape(rev_idx)[0]), rev_idx, tf.shape(res)[0])
        if to_sort:
            linspace = tf.range(tf.shape(count)[0])
            argsorted = tf.argsort(res)
            lookup_table = tf.lookup.StaticHashTable(tf.lookup.KeyValueTensorInitializer(linspace, argsorted),
                                                     default_value=-1)

            rev_idx_sorted = lookup_table.lookup(rev_idx)

            res_sorted = tf.scatter_nd(tf.expand_dims(argsorted, -1), res, tf.shape(res))
            count_sorted = tf.scatter_nd(tf.expand_dims(argsorted, -1), count, tf.shape(res))
            idx_sorted = tf.scatter_nd(tf.expand_dims(argsorted, -1), idx, tf.shape(res))
            return tf.concat([tf.cast(rev_idx_sorted, tf.float32), res_sorted, tf.cast(idx_sorted, tf.float32),
                              tf.cast(count_sorted, tf.float32)],
                             axis=0)
        else:
            return tf.concat([tf.cast(rev_idx, tf.float32), res, tf.cast(idx, tf.float32), tf.cast(count, tf.float32)],
                             axis=0)

    lambda_layer = keras.layers.Lambda(target_layer, name=f"{params['cleaned_name']}_unique")
    lambda_res = lambda_layer(lambda_input)
    rev_idx = lambda_res[:rev_idx_length]
    lambda_length = tf_size(lambda_res, tf_name=f"{params['cleaned_name']}_unique_size_2")
    remainder = tf_cast((lambda_length - rev_idx_length) / 3, tf.int32, tf_name=f"{params['cleaned_name']}_unique_cast1")  # not working need to fix
    count = tf_cast(lambda_res[-remainder:], tf.int32, tf_name=f"{params['cleaned_name']}_unique_cast2")
    idx = tf_cast(lambda_res[-2 * remainder:-remainder], tf.int32, tf_name=f"{params['cleaned_name']}_unique_cast3")
    res = tf_cast(lambda_res[-3 * remainder:-2 * remainder], tf.int32, tf_name=f"{params['cleaned_name']}_unique_cast4")
    layers[keras_name[0]] = res
    layers[keras_name[1]] = idx
    layers[keras_name[2]] = rev_idx
    layers[keras_name[3]] = count



from .convolution_layers import convert_conv, convert_convtranspose
from .activation_layers import convert_relu, convert_elu, convert_lrelu, convert_selu, \
    convert_sigmoid, convert_tanh, convert_softmax, convert_prelu, convert_hard_sigmoid, convert_erf, convert_soft_plus, \
    convert_soft_sign, convert_mish, convert_gelu, convert_hard_swish
from .ltsm_layers import convert_lstm, convert_gru
from .operation_layers import convert_clip, convert_exp, convert_neg, convert_reduce_sum, convert_reduce_mean, \
    convert_log, convert_pow, convert_sqrt, convert_split, convert_cast, convert_floor, convert_identity, \
    convert_argmax, convert_reduce_l2, convert_reduce_max, convert_reciprocal, convert_abs, convert_not, convert_cosine, \
    convert_less, convert_less_equal, convert_and, convert_greater, convert_greater_equal, convert_xor, convert_or, \
    convert_trilu, convert_sign, convert_cosh, convert_sin, convert_sinh, convert_ceil, convert_acosh, convert_acos, \
    convert_asinh, convert_asin, convert_atanh, convert_atan, convert_bitwise_and, convert_argmin, convert_bitwise_xor, \
    convert_bitwise_or, convert_tan, convert_cumsum, convert_bitwise_not, convert_reduce_prod, convert_reduce_min, \
    convert_is_inf, convert_is_nan, convert_size, convert_non_zero, convert_gather_nd, convert_nms, convert_if, \
    convert_einsum
from .elementwise_layers import convert_elementwise_div, convert_elementwise_add, convert_elementwise_mul, \
    convert_elementwise_sub, convert_max, convert_min, convert_mean, convert_equal, convert_where, convert_scatter_nd, \
    convert_round, convert_mod, convert_bitshift
from .linear_layers import convert_gemm, convert_det
from .reshape_layers import convert_transpose, convert_shape, convert_gather, convert_unsqueeze, \
    convert_concat, convert_reshape, convert_flatten, convert_slice, convert_squeeze, convert_expand, convert_resize, \
    convert_tile, convert_gather_elements
from .constant_layers import convert_constant, convert_constant_of_shape, convert_one_hot
from .normalization_layers import convert_batchnorm, convert_instancenorm, convert_dropout, convert_lrn, convert_layernorm
from .pooling_layers import convert_avgpool, convert_global_max_pool, convert_maxpool, convert_global_avg_pool, convert_topk, convert_roi_align
from .padding_layers import convert_padding
from .upsampling_layers import convert_upsample
from .caffe2_layers import convert_alias_with_name, convert_resize_nearest
from .sampling_layers import convert_gridsample, convert_range, convert_unique
from .fft_layers import convert_dft

AVAILABLE_CONVERTERS = {
    'Abs': convert_abs,
    'AliasWithName': convert_alias_with_name,
    'Conv': convert_conv,
    'ConvTranspose': convert_convtranspose,
    'Relu': convert_relu,
    'Resize': convert_resize,
    'Elu': convert_elu,
    'LeakyRelu': convert_lrelu,
    'Sigmoid': convert_sigmoid,
    'HardSigmoid': convert_hard_sigmoid,
    'Tanh': convert_tanh,
    'Selu': convert_selu,
    'Clip': convert_clip,
    'Exp': convert_exp,
    'Neg': convert_neg,
    'Log': convert_log,
    'Softmax': convert_softmax,
    "ScatterND": convert_scatter_nd,
    'PRelu': convert_prelu,
    'ReduceMax': convert_reduce_max,
    'ReduceSum': convert_reduce_sum,
    'ReduceMean': convert_reduce_mean,
    'ReduceProd': convert_reduce_prod,
    'ReduceMin': convert_reduce_min,
    'Pow': convert_pow,
    'Slice': convert_slice,
    'Squeeze': convert_squeeze,
    'Expand': convert_expand,
    'Sqrt': convert_sqrt,
    'Split': convert_split,
    'Cast': convert_cast,
    'Floor': convert_floor,
    'Identity': convert_identity,
    'ArgMax': convert_argmax,
    'ReduceL2': convert_reduce_l2,
    'Max': convert_max,
    'Min': convert_min,
    'Mean': convert_mean,
    'Div': convert_elementwise_div,
    'Add': convert_elementwise_add,
    'Sum': convert_elementwise_add,
    'Mul': convert_elementwise_mul,
    'Sub': convert_elementwise_sub,
    'Gemm': convert_gemm,
    'MatMul': convert_gemm,
    'Transpose': convert_transpose,
    'Constant': convert_constant,
    'BatchNormalization': convert_batchnorm,
    'InstanceNormalization': convert_instancenorm,
    'Dropout': convert_dropout,
    'LRN': convert_lrn,
    'MaxPool': convert_maxpool,
    'GlobalMaxPool': convert_global_max_pool,
    'AveragePool': convert_avgpool,
    'GlobalAveragePool': convert_global_avg_pool,
    'Shape': convert_shape,
    'Gather': convert_gather,
    'Unsqueeze': convert_unsqueeze,
    'Concat': convert_concat,
    'Reshape': convert_reshape,
    'ResizeNearest': convert_resize_nearest,
    'Pad': convert_padding,
    'Flatten': convert_flatten,
    'Upsample': convert_upsample,
    'Erf': convert_erf,
    'Reciprocal': convert_reciprocal,
    'ConstantOfShape': convert_constant_of_shape,
    'Equal': convert_equal,
    'Where': convert_where,
    'LSTM': convert_lstm,
    'GRU': convert_gru,
    'Tile': convert_tile,
    'GridSample': convert_gridsample,
    'Range': convert_range,
    'Not': convert_not,
    'Less': convert_less,
    'Sign': convert_sign,
    'Cosh': convert_cosh,
    'Sin': convert_sin,
    'Sinh': convert_sinh,
    'LessOrEqual': convert_less_equal,
    "And": convert_and,
    "Greater": convert_greater,
    "GreaterOrEqual": convert_greater_equal,
    "Xor": convert_xor,
    "Or": convert_or,
    'Cos': convert_cosine,
    "Trilu": convert_trilu,
    "Ceil": convert_ceil,
    "Acosh": convert_acosh,
    "Acos": convert_acos,
    "Asinh": convert_asinh,
    "Asin": convert_asin,
    "Atanh": convert_atanh,
    "Atan": convert_atan,
    "BitwiseAnd": convert_bitwise_and,
    "BitwiseOr": convert_bitwise_or,
    "BitwiseXor": convert_bitwise_xor,
    "BitwiseNot": convert_bitwise_not,
    "ArgMin": convert_argmin,
    "OneHot": convert_one_hot,
    "Round": convert_round,
    "Tan": convert_tan,
    "CumSum": convert_cumsum,
    "IsInf": convert_is_inf,
    "IsNaN": convert_is_nan,
    "Size": convert_size,
    "Det": convert_det,
    "NonZero": convert_non_zero,
    "GatherND": convert_gather_nd,
    "Softplus": convert_soft_plus,
    "Softsign": convert_soft_sign,
    "Mish": convert_mish,
    "Gelu": convert_gelu,
    "HardSwish": convert_hard_swish,
    "DFT": convert_dft,
    "Mod": convert_mod,
    "BitShift": convert_bitshift,
    "TopK": convert_topk,
    'GatherElements': convert_gather_elements,
    'NonMaxSuppression': convert_nms,
    'Unique': convert_unique,
    'If': convert_if,
    'RoiAlign': convert_roi_align,
    'Einsum': convert_einsum,
    'LayerNormalization': convert_layernorm
}
